from typing import Optional

class weekday(object):
    def __init__(self, weekday: int, n: Optional[int] = ...) -> None: ...

    def __call__(self, n: int) -> weekday: ...

    def __eq__(self, other) -> bool: ...

    def __repr__(self) -> str: ...

    def __hash__(self) -> int: ...

    weekday: int
    n: int



from typing import Callable, List, Optional, Dict, Pattern

class TextWrapper:
    width: int = ...
    initial_indent: str = ...
    subsequent_indent: str = ...
    expand_tabs: bool = ...
    replace_whitespace: bool = ...
    fix_sentence_endings: bool = ...
    drop_whitespace: bool = ...
    break_long_words: bool = ...
    break_on_hyphens: bool = ...
    tabsize: int = ...
    max_lines: Optional[int] = ...
    placeholder: str = ...

    # Attributes not present in documentation
    sentence_end_re: Pattern[str] = ...
    wordsep_re: Pattern[str] = ...
    wordsep_simple_re: Pattern[str] = ...
    whitespace_trans: str = ...
    unicode_whitespace_trans: Dict[int, int] = ...
    uspace: int = ...
    x: int = ...

    def __init__(
            self,
            width: int = ...,
            initial_indent: str = ...,
            subsequent_indent: str = ...,
            expand_tabs: bool = ...,
            replace_whitespace: bool = ...,
            fix_sentence_endings: bool = ...,
            break_long_words: bool = ...,
            drop_whitespace: bool = ...,
            break_on_hyphens: bool = ...,
            tabsize: int = ...,
            *,
            max_lines: Optional[int] = ...,
            placeholder: str = ...) -> None:
        ...

    # Private methods *are* part of the documented API for subclasses.
    def _munge_whitespace(self, text: str) -> str: ...
    def _split(self, text: str) -> List[str]: ...
    def _fix_sentence_endings(self, chunks: List[str]) -> None: ...
    def _handle_long_word(self, reversed_chunks: List[str], cur_line: List[str], cur_len: int, width: int) -> None: ...
    def _wrap_chunks(self, chunks: List[str]) -> List[str]: ...
    def _split_chunks(self, text: str) -> List[str]: ...

    def wrap(self, text: str) -> List[str]: ...
    def fill(self, text: str) -> str: ...


def wrap(
        text: str = ...,
        width: int = ...,
        *,
        initial_indent: str = ...,
        subsequent_indent: str = ...,
        expand_tabs: bool = ...,
        tabsize: int = ...,
        replace_whitespace: bool = ...,
        fix_sentence_endings: bool = ...,
        break_long_words: bool = ...,
        break_on_hyphens: bool = ...,
        drop_whitespace: bool = ...,
        max_lines: int = ...,
        placeholder: str = ...
) -> List[str]:
    ...

def fill(
        text: str,
        width: int = ...,
        *,
        initial_indent: str = ...,
        subsequent_indent: str = ...,
        expand_tabs: bool = ...,
        tabsize: int = ...,
        replace_whitespace: bool = ...,
        fix_sentence_endings: bool = ...,
        break_long_words: bool = ...,
        break_on_hyphens: bool = ...,
        drop_whitespace: bool = ...,
        max_lines: int = ...,
        placeholder: str = ...
) -> str:
    ...

def shorten(
        text: str,
        width: int,
        *,
        initial_indent: str = ...,
        subsequent_indent: str = ...,
        expand_tabs: bool = ...,
        tabsize: int = ...,
        replace_whitespace: bool = ...,
        fix_sentence_endings: bool = ...,
        break_long_words: bool = ...,
        break_on_hyphens: bool = ...,
        drop_whitespace: bool = ...,
        # Omit `max_lines: int = None`, it is forced to 1 here.
        placeholder: str = ...
) -> str:
    ...

def dedent(text: str) -> str:
    ...

def indent(text: str, prefix: str, predicate: Callable[[str], bool] = ...) -> str:
    ...



from fileinput import filelineno
import sys

#We reimplement things from r.py here because then we don't have to import rp, decreasing the delay of the output from running this file.
#This script is called from iterfzf inside of r.py, and might be called tens of times a second


def line_number_prefix_generator( line_number,num_digits):
    return (('%%%ii| ')%num_digits)%line_number 

def line_highlighter(text):
    return fansi((text),'black','bold','yellow')

def preview_line_highlighter(text):
    return fansi((text),'black','bold','green')

def prefix_highlighter(prefix):
    return fansi(prefix,'cyan','bold','black')

def preview_prefix_highlighter(prefix):
    return fansi(prefix,'green','bold','black')

digit_remover=str.maketrans('0123456789', '          ')


def fansi_syntax_highlighting(code: str,
                              namespace=(),
                              style_overrides:dict={},
                              line_wrap_width:int=None,
                              show_line_numbers:bool=False,
                              lazy:bool=False,
                              highlighted_line_numbers=[], #This is not in rp. It's not useful for rp generically, but is very useful for fzf result highlighting...
                              squelch=False,#Supress errors?
                              ):
    #TODO: Because of the way it was programmed, it now included an extraneous new empty line on the top of the output. Feel free to remove that later brutishly lol (just lob it off the final output)
    #If lazy==True, this function returns a generator of strings that should be printed sequentially without new lines
    #If line_wrap_width is an int, it will wrap the whole output to that width - this is suprisingly tricky to do because of the ansi escape codes
    #show_line_numbers, if true, will also display a line number gutter on the side
    #
    #EXAMPLE USING LAZY:
    #    #Lazy can make syntax highlighting of things like rp start instantly
    #    code=get_source_code(r)
    #    for chunk in fansi_syntax_highlighting(code,lazy=True,show_line_numbers=True,line_wrap_width=get_terminal_width()):
    #        print(end=chunk)
    #    print()
    #The result is that it has a shorter delay to start ; but it also might take longer in total
    #
    #EXAMPLE:
    #    print(fansi_syntax_highlighting(get_source_code(load_image),line_wrap_width=30,show_line_numbers=False))
    #
    # PLEASE NOTE THAT I DID NOT WRITE SOME OF THIS CODE!!! IT CAME FROM https://github.com/akheron/cpython/blob/master/Tools/scripts/highlight.py
    # Assumes code was written in python.
    # Method mainly intended for rinsp.
    # I put it in the r class for convenience.
    # Works when I paste methods in but doesn't seem to play nicely with rinsp. I don't know why yet.
    # See the highlight_sourse_in_ansi module for more stuff including HTML highlighting etc.

    highlighted_line_numbers=[x-1 for x in highlighted_line_numbers] #Convert lines 1,2,3 etc to line indices 0,1,2 etc

    default_ansi={
        'comment':('\033[0;31m','\033[0m'),
        'string':('\033[0;32m','\033[0m'),
        'docstring':('\033[0;32m','\033[0m'),
        'keyword':('\033[0;33m','\033[0m'),
        'builtin':('\033[0;35m','\033[0m'),
        'definition':('\033[0;33m','\033[0m'),
        'defname':('\033[0;34m','\033[0m'),
        'operator':('\033[0;33m','\033[0m'),
    }
    default_ansi.update(style_overrides)
    try:
        import keyword,tokenize,re,functools
        try:
            import builtins
        except ImportError:
            import builtins as builtins
        def is_builtin(s):
            'Return True if s is the name of a builtin'
            return hasattr(builtins,s) or s in namespace
        def combine_range(lines,start,end):
            'Join content from a range of lines between start and end'
            (srow,scol),(erow,ecol)=start,end
            if srow == erow:
                return lines[srow - 1][scol:ecol],end
            rows=[lines[srow - 1][scol:]] + lines[srow: erow - 1] + [lines[erow - 1][:ecol]]
            return ''.join(rows),end
        def analyze_python(source):
            '''Generate and classify chunks of Python for syntax highlighting.
               Yields tuples in the form: (category, categorized_text).
            '''
            lines=source.splitlines(True)
            lines.append('')
            readline=functools.partial(next,iter(lines),'')
            kind=tok_str=''
            tok_type=tokenize.COMMENT
            written=(1,0)
            try:
                for tok in tokenize.generate_tokens(readline):
                    prev_tok_type,prev_tok_str=tok_type,tok_str
                    tok_type,tok_str,(srow,scol),(erow,ecol),logical_lineno=tok
                    kind=''
                    if tok_type == tokenize.COMMENT:
                        kind='comment'
                    elif tok_type == tokenize.OP and tok_str[:1] not in '{}[](),.:;@':
                        kind='operator'
                    elif tok_type == tokenize.STRING:
                        kind='string'
                        if prev_tok_type == tokenize.INDENT or scol == 0:
                            kind='docstring'
                    elif tok_type == tokenize.NAME:
                        if tok_str in ('def','class','import','from'):
                            kind='definition'
                        elif prev_tok_str in ('def','class'):
                            kind='defname'
                        elif keyword.iskeyword(tok_str):
                            kind='keyword'
                        elif is_builtin(tok_str) and prev_tok_str != '.':
                            kind='builtin'
                    if kind:
                        if written != (srow,scol):
                            text,written=combine_range(lines,written,(srow,scol))
                            yield '',text
                        text,written=tok_str,(erow,ecol)
                        yield kind,text
            except Exception as e:
                #Probably a syntax error
                # yield from (('',x) for x in readline)
                import traceback
                if not squelch:
                    traceback.print_exception(e)
                pass
            line_upto_token,written=combine_range(lines,written,(erow,ecol))
            yield '',line_upto_token
        def ansi_highlight(classified_text,colors=default_ansi):
            'Add syntax highlighting to source code using ANSI escape sequences'
            # http://en.wikipedia.org/wiki/ANSI_escape_code

            nonlocal line_wrap_width, show_line_numbers

            if line_wrap_width is None:
                line_wrap_width = 9999999
            
            num_code_lines = code.count('\n')+1
            num_digits = len(str(num_code_lines)) #Max number of digits in the line numbers
            num_digits = max(3,num_digits) #Don't jitter around as much in qprp

            max_width = line_wrap_width
            if show_line_numbers:
                #Should always return strings of the same width if done correctly
                #TODO: Make this customizable through the args

                line_prefix_length = len(line_number_prefix_generator(num_code_lines,num_digits))

                if line_wrap_width >= line_prefix_length:
                    max_width = line_wrap_width-line_prefix_length
                else:
                    #We have to not show line numbers, or else we'd be showing literally nothing but them!
                    show_line_numbers=False
                    

            def wrapped_line_tokens(tokens,max_width):
                #Wrap the string, respecting token boundaries when possible
                #Tokens is a list of [(kind,text), (kind,text), ... ] tuples
                #Output is a generator of [(kind,text,line_number) ... ] tuples
                #EXAMPLE TEST:
                #   
                #    >>> list(wrapped_line_tokens([(11,'Hello\nWorld!\n123\nab\nc'),(22,'d'),(33,'e'),(44,'f')],2))
                #    [
                #        (11  , 'He', 0),
                #        (None, '\n', 1),
                #        (11  , 'll', 1),
                #        (None, '\n', 2),
                #        (11  , 'o' , 2),
                #        (None, '\n', 3),
                #        (11  , 'Wo', 3),
                #        (None, '\n', 4),
                #        (11  , 'rl', 4),
                #        (None, '\n', 5),
                #        (11  , 'd!', 5),
                #        (None, '\n', 6),
                #        (11  , '12', 6),
                #        (None, '\n', 7),
                #        (11  , '3' , 7),
                #        (None, '\n', 8),
                #        (11  , 'ab', 8),
                #        (None, '\n', 9),
                #        (11  , 'c' , 9),
                #        (22  , 'd' , 9),
                #        (33  , ''  , 9),
                #        (None, '\n', 10),
                #        (33  , 'e' , 10),
                #        (44  , 'f' , 10)
                #    ]
                #
                line_length=0
                line_number=0
                line_skip=0
                for kind,text in tokens:
                    subtokens=split_including_delimiters(text,'\n')
                    subtokens=subtokens[::-1]
                    while subtokens:
                        assert max_width>=line_length
                        subtoken=subtokens.pop()
                        if subtoken=='\n':
                            if line_number in highlighted_line_numbers:
                                yield 'HIGHLIGHT',' '*1000,line_number #Since FZF doesn't have line wrapping this will go off the edge, highlighting the whole terminal line (assuming the terminal isn't 1000 chars wide lol)
                                # yield 'HIGHLIGHT',' '*(max_width-line_length),line_number 
                            if not line_skip:
                                line_number+=1
                            line_skip=max(0,line_skip-1)
                            line_length=0
                            #Probably can eliminate typehere....
                            yield None,subtoken,line_number
                        elif line_length+len(subtoken)>max_width:
                            index=max_width-line_length
                            token_right=subtoken[index:]
                            subtoken   =subtoken[:index]
                            line_length=0
                            subtokens.append(token_right)
                            subtokens.append('\n')
                            yield kind,subtoken,line_number
                            line_skip+=1
                        else:
                            line_length+=len(subtoken)
                            yield kind,subtoken,line_number
            

            prev_line_number=None
            from itertools import chain
            for kind,text,line_number in chain([[None,'\n',0]],wrapped_line_tokens(classified_text,max_width=max_width)):
                assert isinstance(text,str),'text is not str, is '+str(type(text))
                opener,closer=colors.get(kind,('',''))
                if show_line_numbers and text.endswith('\n'):
                    prefix=line_number_prefix_generator(line_number+1,num_digits)
                    prefix=prefix_highlighter(prefix)
                    # prefix=fansi(prefix,'cyan','bold')#,'black')
                    # prefix='\u001b[1m\u001b[36m'+prefix #cyan bold
                    if line_number==prev_line_number:
                        #https://stackoverflow.com/questions/19084443/replacing-digits-with-str-replace
                        prefix=strip_ansi_escapes(prefix)
                        prefix=prefix.translate(digit_remover)
                        prefix=prefix_highlighter(prefix)
                    if line_number in highlighted_line_numbers:
                        prefix=strip_ansi_escapes(prefix)
                        prefix=line_highlighter(prefix)
                    text=text+prefix
                if line_number in highlighted_line_numbers:
                    yield from line_highlighter(text)
                else:
                    yield from [opener,text,closer]
                prev_line_number=line_number
        output=(ansi_highlight(analyze_python(code)))
        if lazy:
            return output
        else:
            return ''.join(output)
    
    except Exception:
        #To be honest, when the tokenizer throws a hissy fit idk what to do...haven't thought about it too much.
        #It cuts off some of the text in the preview, so that's not good...
        #TODO: figure this out. i dont' have time rn lol.
        #To reproduce this, try doing FDT through a bunch of files - eventually some error will pop up in the preview like "tokenize.TokenError: ('EOF in multi-line statement', (129, 0)) " which refers to the code being parsed
        if not squelch:
            print(code)
            raise
        # return code  # Failed to highlight code, presumably because of an import error.

try:
    import re
    _strip_ansi_escapes=re.compile(r'(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]')
except Exception:pass
def strip_ansi_escapes(string):
    #Undoes anything fansi might do to a string
    #Code from https://www.tutorialspoint.com/How-can-I-remove-the-ANSI-escape-sequences-from-a-string-in-python#targetText=You%20can%20use%20regexes%20to,%5B%40-~%5D'.
    try:
        return _strip_ansi_escapes.sub('',string)
    except NameError:
        assert False,"Failed to import re upon booting rp"


def fansi(text_string,text_color=None,style=None,background_color=None,*,per_line=False,include_end=True):
    #TODO: Fix bug: PROBLEM is that '\n' not in fansi('Hello\n','gray')
    #This function uses ANSI escape sequnces to make colored text in a terminal.
    #It can also make bolded, underlined, or highlighted text.
    #It uses ANSI escape sequences to do this...
    #    ...and so calling it 'fansi' is a pun on 'fancy' and 'ansi'
    # 'fansi' is a pun, referring to ANSI and fancy
    # Uses ANSI formatting to give the terminal color outputs.
    # There are only 8 possible choices from each category, in ［０‚７］⋂ ℤ
    # Adding 0,30,and 40 because of the ANSI codes. Subtracting 1 later on because the syntax
    # of this def says that '0' is the absence of any style etc, whereas 1-8 are active styles.
    # The 'per_line' option applies fansi to every line, which is useful when trying to draw tables and such
    # Some terminals cant handle ansi escape sequences and just print garbage, so if _disable_fansi is turned on this function just returns unformatted text.
    #   (This is usually only the case with more obscure terminals, such as one I have for ssh'ing on my phone. But they do exist)
    # To undo the effect of this function on a string (aka to un-format a string) use rp.strip_ansi_escapes()  (see its documentation for more details)
    # EXAMPLE: print(fansi('ERROR:','red','bold')+fansi(" ATE TOO MANY APPLES!!!",'blue','underlined','yellow'))
    text_string=str(text_string)
    if per_line:
        lines=(text_string).splitlines()
        lines=[fansi(line,text_color,style,background_color,per_line=False) for line in lines]
        return '\n'.join(lines)
    if text_string=='':# Without this, print(fansi("",'blue')+'Hello World'
        return ''
    if isinstance(text_color,str):  # if text_color is a string, convert it into the correct integer and handle the associated exceptions
        text_colors={'black':0,'red':1,'green':2,'yellow':3,'blue':4,'magenta':5,'cyan':6,'gray':7,'grey':7}
        try:
            text_color=text_colors[text_color.lower()]
        except Exception:
            print("ERROR: def fansi: input-error: text_color = '{0}' BUT '{0}' is not a valid key! Replacing text_color as None. Please choose from {1}".format(text_color,str(list(text_colors))))
            text_color=None
    if isinstance(style,str):  # if background_color is a string, convert it into the correct integer
        styles={'bold':1,'faded':2,'underlined':4,'blinking':5,'outlined':7}
        try:
            style=styles[style.lower()]  # I don't know what the other integers do.
        except Exception:
            print("ERROR: def fansi: input-error: style = '{0}' BUT '{0}' is not a valid key! Replacing style as None. Please choose from {1}".format(style,str(list(styles))))
            style=None
    if isinstance(background_color,str):  # if background_color is a string, convert it into the correct integer
        background_colors={'black':0,'red':1,'green':2,'yellow':3,'blue':4,'magenta':5,'cyan':6,'gray':7,'grey':7}
        try:
            background_color=background_colors[background_color.lower()]
        except Exception:
            print("ERROR: def fansi: input-error: background_color = '{0}' BUT '{0}' is not a valid key! Replacing background_color as None. Please choose from {1}".format(background_color,str(list(background_colors))))
            background_color=None

    format=[]
    if style is not None:
        assert 0 <= style <= 7,"style == " + str(style) + " ∴ ¬﹙0 <= style <= 7﹚ ∴ AssertionError"
        style+=0
        format.append(str(style))
    if text_color is not None:
        assert 0 <= text_color <= 7,"text_color == " + str(text_color) + " ∴ ¬﹙0 <= text_color <= 7﹚ ∴ AssertionError"
        text_color+=30
        format.append(str(text_color))
    if background_color is not None:
        assert 0 <= background_color <= 7,"background_color == " + str(background_color) + " ∴ ¬﹙0 <= background_color <= 7﹚ ∴ AssertionError"
        background_color+=40
        format.append(str(background_color))

    if include_end:
        return "\x1b[%sm%s\x1b[0m" % (';'.join(format),str(text_string))  # returns a string with the appropriate formatting applied
    else:
        #Don't end the style...
        return "\x1b[%sm%s" % (';'.join(format),str(text_string))  # returns a string with the appropriate formatting applied



def split_including_delimiters(input: str, delimiter: str):
    """
    Splits an input string, while including the delimiters in the output
    
    Unlike str.split, we can use an empty string as a delimiter
    Unlike str.split, the output will not have any extra empty strings
    Conequently, len(''.split(delimiter))== 0 for all delimiters,
       whereas len(input.split(delimiter))>0 for all inputs and delimiters
    
    INPUTS:
        input: Can be any string
        delimiter: Can be any string

    EXAMPLES:
         >>> split_and_keep_delimiter('Hello World  ! ',' ')
        ans = ['Hello ', 'World ', ' ', '! ', ' ']
         >>> split_and_keep_delimiter("Hello**World**!***", "**")
        ans = ['Hello', '**', 'World', '**', '!', '**', '*']
    EXAMPLES:
        assert split_and_keep_delimiter('-xx-xx-','xx') == ['-', 'xx', '-', 'xx', '-'] # length 5
        assert split_and_keep_delimiter('xx-xx-' ,'xx') == ['xx', '-', 'xx', '-']      # length 4
        assert split_and_keep_delimiter('-xx-xx' ,'xx') == ['-', 'xx', '-', 'xx']      # length 4
        assert split_and_keep_delimiter('xx-xx'  ,'xx') == ['xx', '-', 'xx']           # length 3
        assert split_and_keep_delimiter('xxxx'   ,'xx') == ['xx', 'xx']                # length 2
        assert split_and_keep_delimiter('xxx'    ,'xx') == ['xx', 'x']                 # length 2
        assert split_and_keep_delimiter('x'      ,'xx') == ['x']                       # length 1
        assert split_and_keep_delimiter(''       ,'xx') == []                          # length 0
        assert split_and_keep_delimiter('aaa'    ,'xx') == ['aaa']                     # length 1
        assert split_and_keep_delimiter('aa'     ,'xx') == ['aa']                      # length 1
        assert split_and_keep_delimiter('a'      ,'xx') == ['a']                       # length 1
        assert split_and_keep_delimiter(''       ,''  ) == []                          # length 0
        assert split_and_keep_delimiter('a'      ,''  ) == ['a']                       # length 1
        assert split_and_keep_delimiter('aa'     ,''  ) == ['a', '', 'a']              # length 3
        assert split_and_keep_delimiter('aaa'    ,''  ) == ['a', '', 'a', '', 'a']     # length 5
    """

    # I made this question an answer at https://stackoverflow.com/questions/2136556/in-python-how-do-i-split-a-string-and-keep-the-separators/73562313#73562313

    # Input assertions
    assert isinstance(input,str), "input must be a string but got "+str(input)
    assert isinstance(delimiter,str), "delimiter must be a string but got "+str(input)

    if delimiter:
        # These tokens do not include the delimiter, but are computed quickly
        tokens = input.split(delimiter)
    else:
        # Edge case: if the delimiter is the empty string, split between the characters
        tokens = list(input)
        
    # The following assertions are always true for any string input and delimiter
    # For speed's sake, we disable this assertion
    # assert delimiter.join(tokens) == input

    output = tokens[:1]

    for token in tokens[1:]:
        output.append(delimiter)
        if token:
            output.append(token)
    
    # Don't let the first element be an empty string
    if output[:1]==['']:
        del output[0]
        
    # The only case where we should have an empty string in the output is if it is our delimiter
    # For speed's sake, we disable this assertion
    # assert delimiter=='' or '' not in output
        
    # The resulting strings should be combinable back into the original string
    # For speed's sake, we disable this assertion
    # assert ''.join(output) == input

    return output

# def get_terminal_size():  # In (ↈcolumns，ↈrows) tuple form
#     # From http://stackoverflow.com/questions/566746/how-to-get-linux-console-window-width-in-python/14422538#14422538
#     import os
#     env=os.environ
#     def ioctl_GWINSZ(fd):
#         try:
#             import fcntl,termios,struct,os
#             cr=struct.unpack('hh',fcntl.ioctl(fd,termios.TIOCGWINSZ,
#                                               '1234'))
#         except Exception:
#             return
#         return cr
#     cr=ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)
#     if not cr:
#         try:
#             fd=os.open(os.ctermid(),os.O_RDONLY)
#             cr=ioctl_GWINSZ(fd)
#             os.close(fd)
#         except Exception:
#             pass
#     if not cr:
#         cr=(env.get('LINES',24),env.get('COLUMNS',80))

#         ### Use get(key[, default]) instead of a try/catch
#         # try:
#         #    cr = (env['LINES'], env['COLUMNS'])
#         # except Exception:
#         #    cr = (25, 80)
#     return int(cr[1]),int(cr[0])

# def get_terminal_width():
#     return get_terminal_size()[0]

# def _line_numbered_string(string,foreground='cyan',style='bold',background='blue'):
#     lines=line_split(string)
#     nlines=len(lines)
#     numwidth=len(str(nlines))
#     newlines=[fansi(str(i+1).rjust(numwidth)+' '*0,foreground,style,background)+e for i,e in enumerate(lines)]
#     return line_join(newlines)

# def wrap_string_to_width(string,width):
#     #TODO: Make this work with visible_string_length so that unicode chars/ansi codes are supported
#     #EXAMPLE:
#     # ⮤ wrap_string_to_width('Hello\nWorld!',2)
#     #    ans = He
#     #    ll
#     #    o
#     #    Wo
#     #    rl
#     #    d!
#     assert width>=0,'Cannot have negative width'
#     lines=[]
#     for line in line_split(string):
#         lines+=split_into_sublists(line,width,strict=False,keep_remainder=True)
#     return line_join(''.join(line)for line in lines)

def split_into_sublists(l,sublist_len:int,strict=False,keep_remainder=True):
    # If strict: sublist_len MUST evenly divide len(l)
    # It will return a list of tuples, unless l is a string, in which case it will return a list of strings
    # keep_remainder is not applicable if strict
    # if not keep_remainder and sublist_len DOES NOT evenly divide len(l), we can be sure that all tuples in the output are of len sublist_len, even though the total number of elements in the output is less than in l.
    # EXAMPLES:
    # ⮤ split_into_sublists([1,2,3,4,5,6,7,8,9］,3 ,0)   -> [(1,2,3),(4,5,6),(7,8,9)]
    # ⮤ split_into_sublists([1,2,3,4,5,6,7,8,9］,4 ,0)   -> [(1,2,3,4),(5,6,7,8),(9,)]
    # ⮤ split_into_sublists([1,2,3,4,5,6,7,8,9］,5 ,0)   -> [(1,2,3,4,5),(6,7,8,9)]
    # ⮤ split_into_sublists([1,2,3,4,5,6,7,8,9］,6 ,0)   -> [(1,2,3,4,5,6),(7,8,9)]
    # ⮤ split_into_sublists([1,2,3,4,5,6,7,8,9］,66,0)   -> [(1,2,3,4,5,6,7,8,9)]
    # ⮤ split_into_sublists([1,2,3,4,5,6,7,8,9］,66,0,1) -> [(1,2,3,4,5,6,7,8,9)]
    # ⮤ split_into_sublists([1,2,3,4,5,6,7,8,9］,66,0,0) -> []
    # ⮤ split_into_sublists([1,2,3,4,5,6,7,8,9］,5 ,0,0) -> [(1,2,3,4,5)]
    # ⮤ split_into_sublists([1,2,3,4,5,6,7,8,9］,4 ,0,0) -> [(1,2,3,4),(5,6,7,8)]
    # ⮤ split_into_sublists([1,2,3,4,5,6,7,8,9］,3 ,0,0) -> [(1,2,3),(4,5,6),(7,8,9)]
    # ⮤ split_into_sublists([1,2,3,4,5,6,7,8,9］,4 ,1,0) -> ERROR: ¬ 4 | 9

    assert is_number(sublist_len),'sublist_len should be an integer, but got type '+repr(type(sublist_len))
    if strict:
        assert not len(l)%sublist_len,'len(l)=='+str(len(l))+' and sublist_len=='+str(sublist_len)+': strict mode is turned on but the sublist size doesnt divide the list input evenly. len(l)%sublist_len=='+str(len(l)%sublist_len)+'!=0'
    n=sublist_len

    #This line is rather dense, but it makes sense.
    output=list(zip(*(iter(l),) * n))+([tuple(l[len(l)-len(l)%n:])] if len(l)%n and keep_remainder else [])
    
    if isinstance(l,str):
        output=[''.join(substring) for substring in output]

    return output


def get_absolute_path(path,*,physical=True):
    import os
    #Given a relative path, return its absolute path
    #If physical, expand all symlinks in the path
    path=os.path.expanduser(path)#In case the path has a '~' in it
    if physical:
        path=os.path.realpath(path)#Get rid of any symlinks in the path
    return os.path.abspath(path)

def number_of_lines_in_file(filename):
    #Quickly count the nubmer of lines in a given file.
    #It's 5-10x faster than text_file_to_string(filename).count('\n')
    #It also appears to take constant memory; my memory usage didn't flinch even when I threw a 2gb file at it.
    #Note that it doesn't care if it's a text file or not; it just counts the number of \n bytes in the file!
    #   For example, number_of_lines_in_file('picture.jpg')==280 is a possibility.
    #https://stackoverflow.com/questions/845058/how-to-get-line-count-of-a-large-file-cheaply-in-python
    from itertools import takewhile,repeat
    f = open(filename, 'rb')
    bufgen = takewhile(lambda x: x, (f.raw.read(1024*1024) for _ in repeat(None)))
    return sum( buf.count(b'\n') for buf in bufgen )+1

def file_line_iterator(file_name):
    #Opens a file and iterates through its lines
    #Needs a better name
    file=open(file_name)
    while True:
        line=file.readline()
        if not line:
            return
        #I DONT KNOW WHY THIS ASSERTION DOESN'T ALWAYS WORK BUT SOMETIMES IT FAILS...
        if line.endswith('\n'):
            yield line[:-1]
        else:
            yield line

bytes_per_megabyte=2**20


def total_disc_bytes(path):

    #NOTE: 2**20 bytes is one megabyte

    # path can be either a folder or a file; it will detect that for you. Implemented recursively (checks subfolders)
    # returns total size in bytes
    import os
    def get_file_size(path):
        return os.path.getsize(path)
    def get_folder_size(folder):
        # Get the total disc space of a given directory
        # Source: stackoverflow.com/questions/1392413/calculating-a-directory-size-using-python/1392549
        total_size = get_file_size(folder)
        for item in os.listdir(folder):
            itempath = os.path.join(folder, item)
            if os.path.isfile(itempath):
                total_size += get_file_size(itempath)
            elif os.path.isdir(itempath):
                total_size += get_folder_size(itempath)
        return total_size

    if os.path.isfile(path):
        return get_file_size(path)
    elif os.path.isdir(path):
        return get_folder_size(path)
    else:
        assert False,'r.get_disc_space ERROR: '+path+' is neither a folder nor a file!'

def line_join(x):
    return '\n'.join(x)

def human_readable_file_size(file_size:int):
    #Given a file size in bytes, return a string that represents how large it is in megabytes, gigabytes etc - whatever's easiest to interperet
    #EXAMPLES:
    #     >>> human_readable_file_size(0)
    #    ans = 0B
    #     >>> human_readable_file_size(100)
    #    ans = 100B
    #     >>> human_readable_file_size(1023)
    #    ans = 1023B
    #     >>> human_readable_file_size(1024)
    #    ans = 1KB
    #     >>> human_readable_file_size(1025)
    #    ans = 1.0KB
    #     >>> human_readable_file_size(1000000)
    #    ans = 976.6KB
    #     >>> human_readable_file_size(10000000)
    #    ans = 9.5MB
    #     >>> human_readable_file_size(1000000000)
    #    ans = 953.7MB
    #     >>> human_readable_file_size(10000000000)
    #    ans = 9.3GB
    
    for count in 'B KB MB GB TB PB EB ZB YB BB GB'.split():
        #Bytes Kilobytes Megabytes Gigabytes Terrabytes Petabytes Exobytes Zettabytes Yottabytes Brontobytes Geopbytes
        if file_size > -1024.0 and file_size < 1024.0:
            if int(file_size)==file_size:   
                return "%i%s" % (file_size, count)
            else:
                return "%3.1f%s" % (file_size, count)
        file_size /= 1024.0

import difflib
from typing import Optional

class LineDiffHunk:
    def __init__(self,old_line_number:int,new_line_number:int,type:str):
        assert type in ' -+' and len(type)==1
        self.type=type
        self.lines=[]
        self.old_line_number=old_line_number
        self.new_line_number=new_line_number
    
    @property
    def color(self):    
        return {'+':'green','-':'red',' ':None}[self.type]
    
    def numbered_line(self,line:str,old_line_number:int,new_line_number:int):
        prefix_start=fansi(self.type+' ',self.color,'bold')        
        old_prefix='%4i'%old_line_number
        divider='│'
        new_prefix='%4i'%new_line_number
        prefix_end='│'+prefix_start
        
        if self.type=='+':old_prefix=' '*len(old_prefix)
        if self.type=='-':new_prefix=' '*len(old_prefix)
        
        if self.type!=' ':
            old_prefix=fansi(old_prefix,self.color,'bold')
            new_prefix=fansi(new_prefix,self.color,'bold')
            def fansi_with_underlined_trailing_whitespace(line):
                whitespace_amount=len(line)-len(line.rstrip())
                if whitespace_amount==0:
                    return fansi(line,self.color)
                    
                whitespace=line[-whitespace_amount:]
                return fansi(line[:-whitespace_amount],self.color)+fansi(whitespace,None,None,'yellow')
            line=fansi_with_underlined_trailing_whitespace(line)#,'underlined')
        
        return old_prefix+divider+new_prefix+prefix_end+line

    def __iter__(self):            
        old_line_number=self.old_line_number
        new_line_number=self.new_line_number
        for line in self.lines:
            old_line_number+=1
            new_line_number+=1
            yield self.numbered_line(line,old_line_number,new_line_number)
    
    def __str__(self):
        return line_join(self)
            
    def summary_string(self,max_context=5):#,is_first_hunk=False,is_last_hunk=False):
        cut_line='\n • • • • ( %i lines ) • • • • \n'
        cut_line=fansi(cut_line,'blue')
        
        lines=list(self)
        if len(lines)>2*max_context+1 :#and self.type==' ':
            num_lines_cut=len(lines)-2*max_context
            
            if num_lines_cut>=10:
                #Don't report we just cut one damn line lol...
                output=[]
                output+=lines[:max_context]
                output+=[cut_line%num_lines_cut]
                output+=lines[-max_context:]
                
                lines=output
            
        return line_join(lines)
            

def diff(old: str, new: str):#->list[LineDiffHunk]:

    old = old.splitlines()
    new = new.splitlines()

    line_types = {' ': None, '-': 'red', '+': 'green', '?': 'magenta'}

    old_line_number=0
    new_line_number=0
    

    current_hunk=LineDiffHunk(
        old_line_number,
        new_line_number,
        type=' ',
    )
    
    hunks=[]

    for line in difflib.Differ().compare(old, new):
        
        type=line[0]
        line=line[2:]

        assert type in ' -+?'

        if type=='?':
            # Skip this. See difflib.Differ.__doc__ to see what ? does.
            # It shows where in the line things changed, with an extra line.
            continue
        
        if not current_hunk.lines or type==current_hunk.type:
            current_hunk.type=type
            current_hunk.lines.append(line)
        else:
            assert type!=current_hunk.type
            hunks.append(current_hunk)
            current_hunk=LineDiffHunk(old_line_number,new_line_number,type)
            current_hunk.lines.append(line)

        if type==' ':
            new_line_number+=1
            old_line_number+=1
        elif type=='+':
            new_line_number+=1
        elif type=='-':
            old_line_number+=1
        else:
            assert False,'Unreachable code'
            
        
    hunks.append(current_hunk)
    
    return hunks
            
    
def diff_display_string(a:str,b:str)->str:
    #    #EXAMPLE:
    #        a="# 2022-09-08 00:16:07.547366\nimport difflib\nfrom typing import Optional\n\nclass LineDiffHunk:\n    def __init__(self,old_line_number:int,new_line_number:int,type:str):\n        assert type in ' -+' and len(type)==1\n        self.type=type\n        self.lines=[]\n        self.old_line_number=old_line_number\n        self.new_line_number=new_line_number\n    \n    @property\n    def color(self):    \n        return {'+':'green','-':'red',' ':None}[self.type]\n    \n    def numbered_line(self,line:str,old_line_number:int,new_line_number:int):\n        prefix_start=fansi(self.type+' ',self.color)        \n        old_prefix='%4i'%old_line_number\n        divider=', '\n        new_prefix='%4i'%new_line_number\n        prefix_end='│ '\n        \n        if self.type=='+':old_prefix=' '*len(old_prefix)\n        if self.type=='-':new_prefix=' '*len(old_prefix)\n        \n        new_prefix=fansi(new_prefix,self.color,'bold')\n        old_prefix=fansi(old_prefix,self.color,'bold')\n        \n        return old_prefix+divider+new_prefix+prefix_end+line\n\n    def __iter__(self):            \n        old_line_number=self.old_line_number\n        new_line_number=self.new_line_number\n        for line in self.lines:\n            old_line_number+=1\n            new_line_number+=1\n            yield self.numbered_line(line,old_line_number,new_line_number)\n    \n    def __str__(self):\n        return line_join(self)\n            \n    def summary_string(self,max_context=5,is_first_hunk=False,is_last_hunk=False):\n        cut_line=' • • • • ( %i lines ) • • • • '\n        \n        lines=list(self)\n        if self.type==' ' and len(lines)>2*max_context+1:\n            num_lines_cut=len(lines)-2*max_context\n            self.lines=self.lines[:max_context]+[cut_line%num_lines_cut]+self.lines[-max_context:]\n            \n        return line_join(lines)\n            \n\ndef diff(old: str, new: str)->list[LineDiffHunk]:\n\n    old = old.splitlines()\n    new = new.splitlines()\n\n    line_types = {' ': None, '-': 'red', '+': 'green', '?': 'magenta'}\n\n    old_line_number=0\n    new_line_number=0\n    \n\n    current_hunk=LineDiffHunk(\n        old_line_number,\n        new_line_number,\n        type=' ',\n    )\n    \n    hunks=[]\n\n    for line in difflib.Differ().compare(old, new):\n        \n        type=line[0]\n        line=line[1:]\n\n        assert type in ' -+?'\n        \n        if type==' ':\n            new_line_number+=1\n            old_line_number+=1\n        elif type=='+':\n            new_line_number+=1\n        elif type=='-':\n            old_line_number+=1\n        elif type=='?':\n            # Skip this. See difflib.Differ.__doc__ to see what ? does.\n            # It shows where in the line things changed, with an extra line.\n            continue\n        else:\n            assert False,'Unreachable code'\n            \n        if not current_hunk.lines or type==current_hunk.type:\n            current_hunk.type=type\n            current_hunk.lines.append(line)\n            current_hunk.new_line_number=new_line_number\n            current_hunk.old_line_number=old_line_number            \n        else:\n            assert type!=current_hunk.type\n            hunks.append(current_hunk)\n            current_hunk=LineDiffHunk(old_line_number,new_line_number,type)\n            current_hunk.lines.append(line)\n        \n    hunks.append(current_hunk)\n    \n    return hunks\n            \n    \ndef print_diff(a,b):\n    for hunk in diff(a,b):\n        print(hunk.summary_string())\n"
    #        b="import difflib\nfrom typing import Optional\n\nclass LineDiffHunk:\n    def __init__(self,old_line_number:int,new_line_number:int,type:str):\n        assert type in ' -+' and len(type)==1\n        self.type=type\n        self.lines=[]\n        self.old_line_number=old_line_number\n        self.new_line_number=new_line_number\n    \n    @property\n    def color(self):    \n        return {'+':'green','-':'red',' ':None}[self.type]\n    \n    def numbered_line(self,line:str,old_line_number:int,new_line_number:int):\n        prefix_start=fansi(self.type+' ',self.color,'bold')        \n        old_prefix='%4i'%old_line_number\n        divider='│'\n        new_prefix='%4i'%new_line_number\n        prefix_end='│'+prefix_start\n        \n        if self.type=='+':old_prefix=' '*len(old_prefix)\n        if self.type=='-':new_prefix=' '*len(old_prefix)\n        \n        if self.type!=' ':\n            old_prefix=fansi(old_prefix,self.color,'bold')\n            new_prefix=fansi(new_prefix,self.color,'bold')\n            line=fansi(line,self.color,'underlined')\n        \n        return old_prefix+divider+new_prefix+prefix_end+line\n\n    def __iter__(self):            \n        old_line_number=self.old_line_number\n        new_line_number=self.new_line_number\n        for line in self.lines:\n            old_line_number+=1\n            new_line_number+=1\n            yield self.numbered_line(line,old_line_number,new_line_number)\n    \n    def __str__(self):\n        return line_join(self)\n            \n    def summary_string(self,max_context=5,is_first_hunk=False,is_last_hunk=False):\n        cut_line='\\n • • • • ( %i lines ) • • • • \\n\\n'\n        cut_line=fansi(cut_line,'blue')\n        \n        lines=list(self)\n        if self.type==' ' and len(lines)>2*max_context+1:\n            num_lines_cut=len(lines)-2*max_context\n            \n            output=[]\n            output+=lines[:max_context]\n            output+=[cut_line%num_lines_cut]\n            output+=lines[-max_context:]\n            \n            lines=output\n            \n        return line_join(lines)\n            \n\ndef diff(old: str, new: str)->list[LineDiffHunk]:\n\n    old = old.splitlines()\n    new = new.splitlines()\n\n    line_types = {' ': None, '-': 'red', '+': 'green', '?': 'magenta'}\n\n    old_line_number=0\n    new_line_number=0\n    \n\n    current_hunk=LineDiffHunk(\n        old_line_number,\n        new_line_number,\n        type=' ',\n    )\n    \n    hunks=[]\n\n    for line in difflib.Differ().compare(old, new):\n        \n        type=line[0]\n        line=line[2:]\n\n        assert type in ' -+?'\n\n        if type=='?':\n            # Skip this. See difflib.Differ.__doc__ to see what ? does.\n            # It shows where in the line things changed, with an extra line.\n            continue\n        \n        if not current_hunk.lines or type==current_hunk.type:\n            current_hunk.type=type\n            current_hunk.lines.append(line)\n        else:\n            assert type!=current_hunk.type\n            hunks.append(current_hunk)\n            current_hunk=LineDiffHunk(old_line_number,new_line_number,type)\n            current_hunk.lines.append(line)\n\n        if type==' ':\n            new_line_number+=1\n            old_line_number+=1\n        elif type=='+':\n            new_line_number+=1\n        elif type=='-':\n            old_line_number+=1\n        else:\n            assert False,'Unreachable code'\n            \n        \n    hunks.append(current_hunk)\n    \n    return hunks\n            \n    \ndef print_diff(a,b):\n    for hunk in diff(a,b):\n        print(hunk.summary_string())\n"
    #        print_diff(a,b)
    subs=0
    sames=0
    adds=0
    out=''
    for hunk in diff(a,b):
        if hunk.type==' ':
            sames+=len(hunk.lines)
        if hunk.type=='+':
            adds+=len(hunk.lines)
        if hunk.type=='-':
            subs+=len(hunk.lines)
        out+=(hunk.summary_string())+'\n'
    return subs,sames,adds,out

def split_python_tokens(code: str):
    #Should return a list of all the individual python tokens, INCLUDING whitespace and newlines etc
    #When summed together, the token-strings returned by this function should equal the original inputted string
    from pygments.lexers import Python3Lexer
    from pygments.lexer import Lexer

    def get_all_pygments_tokens(string:str,pygments_lexer:Lexer=Python3Lexer()):
        return pygments_lexer.get_tokens_unprocessed(string)

    def get_all_token_strings(string:str):
        #Returns all the string-value of all tokens parsed from the string, including whitespace and comments
        token_string_generator = (token[2] for token in get_all_pygments_tokens(string))
        return token_string_generator

    return list(get_all_token_strings(code))

def is_valid_python_syntax(code,mode='exec'):
    assert isinstance(code,str),'Code should be a string'
    import ast, traceback
    valid = True
    try:
        ast.parse(code,mode=mode)
    # except SyntaxError: #ValueError: source code string cannot contain null bytes
    except Exception:
        valid = False
    return valid

def print_python_line_summary(code,linum):
    if not is_valid_python_syntax(code):
        return
    linum-=1
    lines=code.splitlines()
    lines=[line+' '*100 for line in lines] #Make sure we still get the path for empty lines with no indent
    def get_indent_level(q):return len(q)-len(q.lstrip())
    current_level=get_indent_level(lines[linum])
    lines=lines[:linum]
    linum=min(linum,len(lines)-1)

    def startswithany(x,things):
        things=things.split()
        x=x.lstrip()
        for y in things:
            if x.startswith(y+' '):
                return True
        return False
    lines=[x for x in lines if startswithany(x,'def class if while for with try except finally async elif else ')]
    levels=[get_indent_level(x) for x in lines]
    lines=lines[::-1]
    levels=levels[::-1]
    out=[]
    for line,level in zip(lines,levels):
        if current_level>level:
            out.append(line)
            current_level=level
            
    out=out[::-1]#Like ['class Rays:', '    def is_cuda(self) -> bool:']
    out=[split_python_tokens(x.strip()) for x in out]
    out=[x[0]+' '+x[2] for x in out]
    def colorize(x):
        if x.startswith('class '):
            return fansi(x,'yellow','bold','black')
        if x.startswith('def '):
            return fansi(x,'green','bold','black')
        return None
        assert False, 'should start with class or def'+ ' '+ x
    out=[colorize(x) for x in out]    
    out=[x for x in out if x is not None]
    out=fansi(' --> ',None,'bold','black').join(out)
    #Out is coloried like "class Rays --> def is_cuda"
    # if not out:
    #     return
    print(fansi('     ',None,None,'black')+out+fansi(' '*1000,None,None,'black'))
    return out

def stdin_read_robust(encoding='utf-8', errors='replace'):
  """
  Reads from standard input robustly by handling potential UnicodeDecodeErrors.

  This function reads the raw byte stream from stdin and decodes it into a 
  string using a specified encoding and error-handling strategy.

  Args:
    encoding (str): The text encoding to use (e.g., 'utf-8', 'latin-1'). 
                    Defaults to 'utf-8'.
    errors (str): The strategy for handling decoding errors. Common options are
                  'replace' (inserts a placeholder ), 'ignore' (discards the
                  problematic byte), or 'strict' (raises an error). 
                  Defaults to 'replace'.

  Returns:
    str: The decoded string from standard input.
  """
  # Read the raw, undecoded bytes from the standard input buffer
  raw_input = sys.stdin.buffer.read()
  
  # Decode the bytes into a string using the specified error handling
  text_input = raw_input.decode(encoding, errors=errors)
  
  return text_input

def start(text=None):

    # HARDCODED_WIDTH=None

    text=text or stdin_read_robust()

    lines=text.splitlines()
    if len(lines)==1 and '\\n' in lines[0]:
        #Some glitch idk why makes it a single line in tmux on sensei
        lines=lines[0].split('\\n')
        text='\n'.join(lines)

    width=int(sys.argv[1])

    text=text.replace('\\"','"')


    try:
        thingstoprint=None
        mode=None
        if len(sys.argv)>2:
            mode = sys.argv[2]

            if mode=='FDT':
                #Something like
                #w​e​b​_​e​v​a​l​u​a​t​o​r​.​p​y​:​1​4​4​:​     def evaluate(self,code:str='',**vars):
                file,line_number,line=text.split(':',maxsplit=2)
                file=file.replace('\u200b','')
                line_number=line_number.replace('\u200b','')
                line_number=int(line_number)
                file_size=total_disc_bytes(file)
                total_num_lines=number_of_lines_in_file(file)



                print(fansi('Line %4i / %i (%s):  '%(line_number,total_num_lines,human_readable_file_size(file_size)),'green','bold','black')+fansi((file),'green','underlined','black')+fansi(' '*1000,None,None,'black'))


                max_megabytes=1
                max_mini_preview_megabytes=256 #a wild guess how big is too big...feel free to calibrate these max bmegabyte numbers Based on your preferences
                max_highlightable_size=bytes_per_megabyte*max_megabytes
                max_minipreviewable_size=bytes_per_megabyte*max_mini_preview_megabytes

                def make_preview_window():
                    num_surrounding_lines=10 #How much context to give in the preview?

                    blank_line='·'*1000

                    def get_preview_line(linenum):
                        linenum+=1
                        prefix=line_number_prefix_generator(linenum+1,num_digits)
                        if 0<=linenum<len(file_lines):
                            line=file_lines[linenum]
                            if linenum==line_number-1:
                                line=preview_line_highlighter(line+' '*1000)
                            else:
                                try:
                                    line=fansi_syntax_highlighting(line,squelch=True)[1:]
                                except Exception as e:
                                    pass
                        else:
                            prefix=prefix.translate(digit_remover).replace('-',' ')
                            line=blank_line
                        if linenum!=line_number-1:
                            prefix=preview_prefix_highlighter(prefix)
                        else:
                            prefix=preview_line_highlighter(prefix)
                        line=prefix+line
                        return line

                    for i in range(line_number-num_surrounding_lines-2,line_number+num_surrounding_lines-1):

                        print(get_preview_line(i))

                    print(prefix_highlighter('    Full Source Code (scroll down):'+' '*1000))

                    line_number

                if file_size>max_highlightable_size:
                    print(fansi('  - NOTE: Syntax highlighting and line wrapping disabled because this file is larger than %i megabytes (it would be annoyingly slow)'%max_megabytes+' '*1000,'green',None,'black'))

                    line_iterator=file_line_iterator(file) #Worst case if its too damn big...


                    if file_size<max_minipreviewable_size:
                        file_text=open(file,'r').read()
                        file_lines=file_text.splitlines()
                        line_iterator=file_lines
                        num_digits=len(str(total_num_lines))
                        num_digits=max(num_digits,4)
                        make_preview_window()
                    else:
                        print(fansi('  - NOTE: Mini-preview disabled because this file is larger than %i megabytes'%max_mini_preview_megabytes+' '*1000,'green',None,'black'))

                    

                    for i,line in enumerate(line_iterator):
                        i+=1
                        num_digits=len(str(total_num_lines))
                        num_digits=max(num_digits,4)
                        prefix=line_number_prefix_generator(i,num_digits)
                        prefix=prefix_highlighter(prefix)
                        print(end=prefix)
                        if i==line_number:
                            line+=' '*1000
                            line=line_highlighter(line)
                        print(line)

                else:

                    num_digits=len(str(total_num_lines))
                    num_digits=max(num_digits,3)
                    file_text=open(file,'r').read()
                    file_lines=file_text.splitlines()
                    
                    try:
                        print_python_line_summary('\n'.join(file_lines),line_number)
                    except Exception:
                        pass
                    make_preview_window()

                    thingstoprint=fansi_syntax_highlighting(file_text,show_line_numbers=True,line_wrap_width=width,lazy=True,highlighted_line_numbers=[line_number])
                # from random import random
                # if random()>.99:
                    # while True:pass

            if mode=='diff_mode':
                import json
                # text=json.loads(text)
                old_code = sys.argv[3]
                old_code=json.loads(old_code)
                # old_code = old_code.replace('\\"','"')
                subs,sames,adds,diff_string=diff_display_string(old_code,text)
                print(fansi(' DIFF FROM SELECTED TO ORIGINAL: '+' '*10000,'cyan','bold','black'))

                print(end=fansi('   LINES: ','cyan','bold','black'))
                sames=fansi('%i untouched, '%sames,'gray','bold','black')
                subs=fansi('  -%i'%subs,'red' ,'bold','black')+fansi(' removed, ','red'  ,'bold','black')
                adds=fansi('  +%i'%adds,'green','bold','black')+fansi(' added','green','bold','black')
                print(sames+subs+adds+fansi(' '*10000,'cyan','bold','black'))
                if text==old_code:
                    print(fansi('          (They\'re identical!) '+' '*10000,'cyan','bold','black'))
                print(end=diff_string)
                

                # print(old_code)

                # text.replace('\\"','"')
                # print("DONE!!!")
                print()
                print(fansi(' SELECTED CODE:  %i line%s'%(text.count('\n')+1,'s' if text.count('\n')+1>1 else '')+' '*10000,'cyan','bold','black'))

                import itertools
                thingstoprint=itertools.chain(
                    fansi_syntax_highlighting(text,show_line_numbers=True,line_wrap_width=width,lazy=True),
                    [' \n'],
                    [' \n'],
                    [fansi(' ORIGINAL CODE:  %i line%s'%(old_code.count('\n')+1,'s' if old_code.count('\n')+1>1 else '')+' '*10000,'cyan','bold','black')],
                    fansi_syntax_highlighting(old_code,show_line_numbers=True,line_wrap_width=width,lazy=True),
                )


        else:
        # width=HARDCODED_WIDTH or 80
        # text=fansi_syntax_highlighting(text)
        # print(text)
        # text=_line_numbered_string(text)
            thingstoprint=fansi_syntax_highlighting(text,show_line_numbers=True,line_wrap_width=width,lazy=True)
        
        if thingstoprint is not None:
            thingstoprint=iter(thingstoprint)
            print(end=next(thingstoprint).replace('\n',''))#Get rid of the first newline - its junk. Maybe a newer version of rp.fansi_syntax_highlighting addresses this, but in this file it outputs an empty junk line at the begginnging.

            print(end=next(thingstoprint).replace('\n','',1))#Get rid of the first newline - its junk. Maybe a newer version of rp.fansi_syntax_highlighting addresses this, but in this file it outputs an empty junk line at the begginnging.
            flag=9#This is a janky hack to prevent it from printing junk on the first line when the first line is selected...idk why it works...
            for x in thingstoprint:
                if mode=='FDT' and flag and line_number==1:
                    flag-=1
                    x=x.replace('\n','')
                print(end=x)
            print()

        exit()
    except Exception as e:
        import traceback
        traceback.print_exception(e)
        # print(e)


start()



from .ansi import ANSIMethod
from .method import RenderMethod


class Ansi8FblockMethod(RenderMethod):
  def __init__(self, image):
    super().__init__(image)

  def to_string(self):
    return ANSIMethod(self.image).to_8_bit_fblock()

class Ansi24FblockMethod(RenderMethod):
  def __init__(self, image):
    super().__init__(image)

  def to_string(self):
    return ANSIMethod(self.image).to_24_bit_fblock()

class Ansi8HblockMethod(RenderMethod):
  def __init__(self, image):
    super().__init__(image)

  def to_string(self):
    return ANSIMethod(self.image).to_8_bit_hblock()

class Ansi24HblockMethod(RenderMethod):
  def __init__(self, image):
    super().__init__(image)

  def to_string(self):
    return ANSIMethod(self.image).to_24_bit_hblock()



from . import client
from . import connection
from . import utils
from . import exceptions

Redis = client.Redis
StrictRedis = client.StrictRedis
BlockingConnectionPool = connection.BlockingConnectionPool
ConnectionPool = connection.ConnectionPool
Connection = connection.Connection
SSLConnection = connection.SSLConnection
UnixDomainSocketConnection = connection.UnixDomainSocketConnection
from_url = utils.from_url
AuthenticationError = exceptions.AuthenticationError
BusyLoadingError = exceptions.BusyLoadingError
ConnectionError = exceptions.ConnectionError
DataError = exceptions.DataError
InvalidResponse = exceptions.InvalidResponse
PubSubError = exceptions.PubSubError
ReadOnlyError = exceptions.ReadOnlyError
RedisError = exceptions.RedisError
ResponseError = exceptions.ResponseError
TimeoutError = exceptions.TimeoutError
WatchError = exceptions.WatchError



__revision__: str

class BaseRNG:
    closed: bool
    def __init__(self) -> None: ...
    def __del__(self): ...
    def __enter__(self): ...
    def __exit__(self): ...
    def close(self): ...
    def flush(self): ...
    def read(self, N: int = ...): ...



import sys

from typing import Any, Callable, Dict, Iterable, List, Optional, Sequence, Text, Tuple, Union
from markupsafe._compat import text_type
import string
from markupsafe._native import escape as escape, escape_silent as escape_silent, soft_unicode as soft_unicode

try:
    from collections.abc import Mapping #Python 3.9 and below
except ImportError:
    from collections import Mapping #Python 3.10+

class Markup(text_type):
    def __new__(cls, base: Text = ..., encoding: Optional[Text] = ..., errors: Text = ...) -> Markup: ...
    def __html__(self) -> Markup: ...
    def __add__(self, other: text_type) -> Markup: ...
    def __radd__(self, other: text_type) -> Markup: ...
    def __mul__(self, num: int) -> Markup: ...
    def __rmul__(self, num: int) -> Markup: ...
    def __mod__(self, *args: Any) -> Markup: ...
    def join(self, seq: Iterable[text_type]): ...
    def split(self, sep: Optional[text_type] = ..., maxsplit: int = ...) -> List[text_type]: ...
    def rsplit(self, sep: Optional[text_type] = ..., maxsplit: int = ...) -> List[text_type]: ...
    def splitlines(self, keepends: bool = ...) -> List[text_type]: ...
    def unescape(self) -> Text: ...
    def striptags(self) -> Text: ...
    @classmethod
    def escape(cls, s: text_type) -> Markup: ...
    def partition(self, sep: text_type) -> Tuple[Markup, Markup, Markup]: ...
    def rpartition(self, sep: text_type) -> Tuple[Markup, Markup, Markup]: ...
    def format(self, *args, **kwargs) -> Markup: ...
    def __html_format__(self, format_spec) -> Markup: ...
    def __getslice__(self, start: int, stop: int) -> Markup: ...
    def __getitem__(self, i: Union[int, slice]) -> Markup: ...
    def capitalize(self) -> Markup: ...
    def title(self) -> Markup: ...
    def lower(self) -> Markup: ...
    def upper(self) -> Markup: ...
    def swapcase(self) -> Markup: ...
    def replace(self, old: text_type, new: text_type, count: int = ...) -> Markup: ...
    def ljust(self, width: int, fillchar: text_type = ...) -> Markup: ...
    def rjust(self, width: int, fillchar: text_type = ...) -> Markup: ...
    def lstrip(self, chars: Optional[text_type] = ...) -> Markup: ...
    def rstrip(self, chars: Optional[text_type] = ...) -> Markup: ...
    def strip(self, chars: Optional[text_type] = ...) -> Markup: ...
    def center(self, width: int, fillchar: text_type = ...) -> Markup: ...
    def zfill(self, width: int) -> Markup: ...
    def translate(self, table: Union[Mapping[int, Union[int, text_type, None]], Sequence[Union[int, text_type, None]]]) -> Markup: ...
    def expandtabs(self, tabsize: int = ...) -> Markup: ...

class EscapeFormatter(string.Formatter):
    escape: Callable[[text_type], Markup]
    def __init__(self, escape: Callable[[text_type], Markup]) -> None: ...
    def format_field(self, value: text_type, format_spec: text_type) -> Markup: ...

if sys.version_info[0] >= 3:
    soft_str = soft_unicode



class _PreferencesMixin:
    _color_scheme = {
        "error": "red",
        "callable_": "cyan",
        "attr": "magenta",
        "message": "white",
        "null": "grey",
        "forged": "yellow",
        "specified": "orange",
        "default": "grey",
        "None": "grey",
        "builtin": "grey",
    }
    _display_type_annotations = True

    @classmethod
    def _load_preferences(cls):
        pass

    @classmethod
    def _update_color_scheme(cls, key, value):
        cls._color_scheme[key] = value



import inspect
import atexit
import rp

#TODO: Integrate this into RP's repl too somehow
#      when repl command finishes, we should also trigger it - not just atexit
#      Also we should be able to do this when using <c-e> -- should modify the buffer (can be undone with <c-z> as normal)
#      SHould be able to toggle this on and off globally to disable debug_eval's
#      Should be able to make it verbose or silent
#      Perhaps ask a yes/no at the end of whether we shoudl replace
#      Inform the user of what files have changed at least...ONLY IF they changed...
#      Allow a frames_back or similar option to debug_comment so other functions can build upon it, adding defaults etc.

def is_zero_arg_callable(func):
    if callable(func):
        if not inspect.signature(func).parameters:
            return True
    return False

# List to store results and metadata
_debug_comment_results = []

def debug_comment(value):
    """
    Modifies the source code of the file this was called in!
    Best practice is to commit all changed to git before this is called so you can revert changes.

    Notes:
        Because it simply adds or modifies # comments, the number of lines in a modified file will not be changed by this code

    EXAMPLE:
        >>> import rp

        >>> #This function modifies your code - it adds comments next to debug_comment that show what it evaluated
        >>> rp.debug_comment(rp.debug_comment)# --> <function debug_comment at 0x10717cee0>

        >>> #If you align the comments like this it will respect the alignment upon updating
        >>> rp.debug_comment(rp.debug_comment.__module__)# --> rp.libs.debug_comment
        >>> rp.debug_comment(rp)                         # --> <module 'rp' from '/opt/homebrew/lib/python3.10/site-packages/rp/__init__.py'>
        >>> rp.debug_comment(__file__)                   # --> /Users/ryan/CleanCode/Sandbox/debugevaltest/evaltest.py
        >>> rp.debug_comment(__name__)                   # --> __main__

        >>> rp.debug_comment("What\nHappens\nWith\nMultiple\nLines")# --> 'What\nHappens\nWith\nMultiple\nLines'

        >>> #The last evaluation will be the one that's kept
        >>> for i in range(5):
        >>> rp.debug_comment(i)   # --> 4
        >>> rp.debug_comment(i**2)# --> 16

        >>> rp.debug_comment(# --> I is 4
        >>> "I is "+str(i)
        >>> )   

        >>> #This is what happens if you call it on multiple lines
        >>> rp.debug_comment(# --> 127
        >>> 123+i
        >>> )   

        >>> #If you give it a function that takes 0 args, it will try to evaluate it
        >>> #This is good for if you want to disable debug_comments and disable all the evaluation inside them too to make your code faster
        >>> rp.debug_comment(lambda: 22*22)# --> 484

        >>> #Use cases:
        >>> import torch
        >>> t=torch.randn(10,100).cpu()
        >>> rp.debug_comment(t.shape) # --> torch.Size([10, 100])
        >>> rp.debug_comment(t.dtype) # --> torch.float32
        >>> rp.debug_comment(t.min()) # --> tensor(-3.1440)
        >>> rp.debug_comment(t.max()) # --> tensor(2.9910)
        >>> rp.debug_comment(t.device)# --> cpu
    """
    # TODO: allow varargs
    # Kwargs will be reserved for special purposes, like if we want to also record the date, or number of times the value changed or number of times we called debug comment etc etc

    # Get caller's frame
    frame = inspect.currentframe().f_back
    filename = frame.f_globals['__file__']
    lineno = frame.f_lineno
    
    try:
        if is_zero_arg_callable(value):
            # If value is a zero-argument callable, call it
            result = value()
        ## Disabled eval-strings because I found I often feed it strings...this only makes sense if we give it a string literal, requiring complex introspection and parsing
        # elif isinstance(value, str):
        #     # Evaluate the string expression
        #     result = eval(value, frame.f_globals, frame.f_locals)
        else:
            # For non-callable, non-string values, convert directly to string
            result = value

        try:
            result = str(result)
        except Exception as e:
            result = "Error turning value into string: " + str(e)

    except Exception as e:
        result = "Error: " + str(e)

    if '\n' in result or '\r' in result: 
        result=repr(result) #No multiline allowed! Has to fit in a single comment string

    _debug_comment_results.append((filename, lineno, result))
    
    #Let original value pass through
    return value

def format_debug_comment(result):
    # _debug_print(result) # This can get a bit spammy...
    return comment_prefix+str(result)

def _debug_print(x):
    rp.fansi_print('rp.debug_comment: '+str(x), 'green', 'underlined')

comment_prefix = "# --> " #Make this regex-friendly. TODO: regex escape this

def _perform_debug_eval_substitutions():
    # Process each file only once
    file_changes = {}
    for filename, lineno, result in _debug_comment_results:
        if filename not in file_changes:
            with open(filename, 'r') as file:
                file_changes[filename] = file.readlines()
        
        # Prepare new comment
        new_comment = format_debug_comment(result)
        current_line = file_changes[filename][lineno - 1]
        
        #TODO: Do this with split_python_tokens instead and do more checks to make sure the source wasn't modified as it was running so that we're replacing the right line. AKA check to make sure this line in the file matches the line in the linecache, and starts with either rp.debug_comment or debug_comment. And check to make sure we only modify the last --> because --> might be in the value itself (unlikely edge case granted).

        # Check if there's an existing debug comment and replace it
        if comment_prefix in current_line:
            current_line = current_line[:current_line.rfind(comment_prefix)] + new_comment + '\n'
        else:
            current_line = current_line.rstrip() + new_comment + '\n'
        
        file_changes[filename][lineno - 1] = current_line

    # Write changes back to the files
    for filename, lines in file_changes.items():
        with open(filename, 'w') as file:
            try:
                _debug_print("writing to "+filename)
                file.writelines(lines)
            except Exception:
                _debug_print("Failed to modify "+filename)

    #Don't do anything twice
    _debug_comment_results.clear()

atexit.register(_perform_debug_eval_substitutions)

# Example usage within the same file
if __name__ == "__main__":
    x = 10
    debug_comment(lambda:'#####'+ str(x * 20))  # This should replace or append the result `20` to this line in the source file. # --> #####200# ----> #####200
    for _ in range(10):
        from rp import random_int
        debug_comment(lambda:"What is "+str(344*random_int(0,100))) # --> What is 11008# ----> What is 24424



import pathlib

import dacite
from loguru import logger
from overrides import overrides

from ...config.config_transformer import AbstractConfigurationTransformer
from ...config.config_error import ConfigurationError

from .keymapper import KeyMapper


def create_keymapper(mode):
    try:
        return KeyMapperConfigurationParser.transform(mode)
    except Exception as e:
        logger.opt(exception=True).error(e)
        raise ConfigurationError(
            f"Failed to load key-mapping from mode name {mode}."
            f"Please consider create an issue at "
            f"https://github.com/cielong/pyfx/issues."
        )


class KeyMapperConfigurationParser(AbstractConfigurationTransformer):
    """ Configuration parser that create keymappers from configuration """

    __HERE = pathlib.Path(__file__).parent.resolve()
    MODES = {
        "basic": None,
        "emacs": __HERE / "modes" / "emacs.yml",
        "vim": __HERE / "modes" / "vim.yml"
    }

    @classmethod
    @overrides(check_signature=False)
    def transform(cls, mode, *args, **kwargs):
        mode_config = cls.MODES[mode]
        if mode_config is None:
            return KeyMapper()

        return dacite.from_dict(
            data_class=KeyMapper, data=cls.load_yaml(mode_config)
        )



# Stubs for six.moves.urllib.response
#
# Note: Commented out items means they weren't implemented at the time.
# Uncomment them when the modules have been added to the typeshed.
# from urllib.response import addbase as addbase
# from urllib.response import addclosehook as addclosehook
# from urllib.response import addinfo as addinfo
from urllib.response import addinfourl as addinfourl



"""
User interface Controls for the layout.
"""
from __future__ import unicode_literals

from abc import ABCMeta, abstractmethod
from collections import namedtuple
from six import with_metaclass
from six.moves import range

from rp.prompt_toolkit.cache import SimpleCache
from rp.prompt_toolkit.enums import DEFAULT_BUFFER, SEARCH_BUFFER
from rp.prompt_toolkit.filters import to_cli_filter
from rp.prompt_toolkit.mouse_events import MouseEventType
from rp.prompt_toolkit.search_state import SearchState
from rp.prompt_toolkit.selection import SelectionType
from rp.prompt_toolkit.token import Token
from rp.prompt_toolkit.utils import get_cwidth

from .lexers import Lexer, SimpleLexer
from .processors import Processor
from .screen import Char, Point
from .utils import token_list_width, split_lines, token_list_to_text

import six
import time


__all__ = (
    'BufferControl',
    'FillControl',
    'TokenListControl',
    'UIControl',
    'UIContent',
)


class UIControl(with_metaclass(ABCMeta, object)):
    """
    Base class for all user interface controls.
    """
    def reset(self):
        # Default reset. (Doesn't have to be implemented.)
        pass

    def preferred_width(self, cli, max_available_width):
        return None

    def preferred_height(self, cli, width, max_available_height, wrap_lines):
        return None

    def has_focus(self, cli):
        """
        Return ``True`` when this user control has the focus.

        If so, the cursor will be displayed according to the cursor position
        reported by :meth:`.UIControl.create_content`. If the created content
        has the property ``show_cursor=False``, the cursor will be hidden from
        the output.
        """
        return False

    @abstractmethod
    def create_content(self, cli, width, height):
        """
        Generate the content for this user control.

        Returns a :class:`.UIContent` instance.
        """

    def mouse_handler(self, cli, mouse_event):
        """
        Handle mouse events.

        When `NotImplemented` is returned, it means that the given event is not
        handled by the `UIControl` itself. The `Window` or key bindings can
        decide to handle this event as scrolling or changing focus.

        :param cli: `CommandLineInterface` instance.
        :param mouse_event: `MouseEvent` instance.
        """
        return NotImplemented

    def move_cursor_down(self, cli):
        """
        Request to move the cursor down.
        This happens when scrolling down and the cursor is completely at the
        top.
        """

    def move_cursor_up(self, cli):
        """
        Request to move the cursor up.
        """


class UIContent(object):
    """
    Content generated by a user control. This content consists of a list of
    lines.

    :param get_line: Callable that returns the current line. This is a list of
        (Token, text) tuples.
    :param line_count: The number of lines.
    :param cursor_position: a :class:`.Point` for the cursor position.
    :param menu_position: a :class:`.Point` for the menu position.
    :param show_cursor: Make the cursor visible.
    :param default_char: The default :class:`.Char` for filling the background.
    """
    def __init__(self, get_line=None, line_count=0,
                 cursor_position=None, menu_position=None, show_cursor=True,
                 default_char=None):
        assert callable(get_line)
        assert isinstance(line_count, six.integer_types)
        assert cursor_position is None or isinstance(cursor_position, Point)
        assert menu_position is None or isinstance(menu_position, Point)
        assert default_char is None or isinstance(default_char, Char)

        self.get_line = get_line
        self.line_count = line_count
        self.cursor_position = cursor_position or Point(0, 0)
        self.menu_position = menu_position
        self.show_cursor = show_cursor
        self.default_char = default_char

        # Cache for line heights. Maps (lineno, width) -> height.
        self._line_heights = {}

    def __getitem__(self, lineno):
        " Make it iterable (iterate line by line). "
        if lineno < self.line_count:
            return self.get_line(lineno)
        else:
            raise IndexError

    def get_height_for_line(self, lineno, width):
        """
        Return the height that a given line would need if it is rendered in a
        space with the given width.
        """
        try:
            return self._line_heights[lineno, width]
        except KeyError:
            try:
                text = token_list_to_text(self.get_line(lineno))
                result = self.get_height_for_text(text, width)

                # Cache and return
                self._line_heights[lineno, width] = result
                return result
            except Exception:
                # If we encounter any errors (such as invalid line numbers),
                # return a default height of 1 and cache it
                self._line_heights[lineno, width] = 1
                return 1

    @staticmethod
    def get_height_for_text(text, width):
        # Get text width for this line.
        line_width = get_cwidth(text)

        # Calculate height.
        try:
            quotient, remainder = divmod(line_width, width)
        except ZeroDivisionError:
            # Return something very big.
            # (This can happen, when the Window gets very small.)
            return 10 ** 10
        else:
            if remainder:
                quotient += 1  # Like math.ceil.
            return max(1, quotient)


class TokenListControl(UIControl):
    """
    Control that displays a list of (Token, text) tuples.
    (It's mostly optimized for rather small widgets, like toolbars, menus, etc...)

    Mouse support:

        The list of tokens can also contain tuples of three items, looking like:
        (Token, text, handler). When mouse support is enabled and the user
        clicks on this token, then the given handler is called. That handler
        should accept two inputs: (CommandLineInterface, MouseEvent) and it
        should either handle the event or return `NotImplemented` in case we
        want the containing Window to handle this event.

    :param get_tokens: Callable that takes a `CommandLineInterface` instance
        and returns the list of (Token, text) tuples to be displayed right now.
    :param default_char: default :class:`.Char` (character and Token) to use
        for the background when there is more space available than `get_tokens`
        returns.
    :param get_default_char: Like `default_char`, but this is a callable that
        takes a :class:`prompt_toolkit.interface.CommandLineInterface` and
        returns a :class:`.Char` instance.
    :param has_focus: `bool` or `CLIFilter`, when this evaluates to `True`,
        this UI control will take the focus. The cursor will be shown in the
        upper left corner of this control, unless `get_token` returns a
        ``Token.SetCursorPosition`` token somewhere in the token list, then the
        cursor will be shown there.
    """
    def __init__(self, get_tokens, default_char=None, get_default_char=None,
                 align_right=False, align_center=False, has_focus=False):
        assert callable(get_tokens)
        assert default_char is None or isinstance(default_char, Char)
        assert get_default_char is None or callable(get_default_char)
        assert not (default_char and get_default_char)

        self.align_right = to_cli_filter(align_right)
        self.align_center = to_cli_filter(align_center)
        self._has_focus_filter = to_cli_filter(has_focus)

        self.get_tokens = get_tokens

        # Construct `get_default_char` callable.
        if default_char:
            get_default_char = lambda _: default_char
        elif not get_default_char:
            get_default_char = lambda _: Char(' ', Token.Transparent)

        self.get_default_char = get_default_char

        #: Cache for the content.
        self._content_cache = SimpleCache(maxsize=18)
        self._token_cache = SimpleCache(maxsize=1)
            # Only cache one token list. We don't need the previous item.

        # Render info for the mouse support.
        self._tokens = None

    def reset(self):
        self._tokens = None

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.get_tokens)

    def _get_tokens_cached(self, cli):
        """
        Get tokens, but only retrieve tokens once during one render run.
        (This function is called several times during one rendering, because
        we also need those for calculating the dimensions.)
        """
        return self._token_cache.get(
            cli.render_counter, lambda: self.get_tokens(cli))

    def has_focus(self, cli):
        return self._has_focus_filter(cli)

    def preferred_width(self, cli, max_available_width):
        """
        Return the preferred width for this control.
        That is the width of the longest line.
        """
        text = token_list_to_text(self._get_tokens_cached(cli))
        line_lengths = [get_cwidth(l) for l in text.split('\n')]
        return max(line_lengths)

    def preferred_height(self, cli, width, max_available_height, wrap_lines):
        content = self.create_content(cli, width, None)
        return content.line_count

    def create_content(self, cli, width, height):
        # Get tokens
        tokens_with_mouse_handlers = self._get_tokens_cached(cli)

        default_char = self.get_default_char(cli)

        # Wrap/align right/center parameters.
        right = self.align_right(cli)
        center = self.align_center(cli)

        def process_line(line):
            " Center or right align a single line. "
            used_width = token_list_width(line)
            padding = width - used_width
            if center:
                padding = int(padding / 2)
            return [(default_char.token, default_char.char * padding)] + line

        if right or center:
            token_lines_with_mouse_handlers = []

            for line in split_lines(tokens_with_mouse_handlers):
                token_lines_with_mouse_handlers.append(process_line(line))
        else:
            token_lines_with_mouse_handlers = list(split_lines(tokens_with_mouse_handlers))

        # Strip mouse handlers from tokens.
        token_lines = [
            [tuple(item[:2]) for item in line]
            for line in token_lines_with_mouse_handlers
        ]

        # Keep track of the tokens with mouse handler, for later use in
        # `mouse_handler`.
        self._tokens = tokens_with_mouse_handlers

        # If there is a `Token.SetCursorPosition` in the token list, set the
        # cursor position here.
        def get_cursor_position():
            SetCursorPosition = Token.SetCursorPosition

            for y, line in enumerate(token_lines):
                x = 0
                for token, text in line:
                    if token == SetCursorPosition:
                        return Point(x=x, y=y)
                    x += len(text)
            return None

        # Create content, or take it from the cache.
        key = (default_char.char, default_char.token,
                tuple(tokens_with_mouse_handlers), width, right, center)

        def get_content():
            return UIContent(get_line=lambda i: token_lines[i],
                             line_count=len(token_lines),
                             default_char=default_char,
                             cursor_position=get_cursor_position())

        return self._content_cache.get(key, get_content)

    @classmethod
    def static(cls, tokens):
        def get_static_tokens(cli):
            return tokens
        return cls(get_static_tokens)

    def mouse_handler(self, cli, mouse_event):
        """
        Handle mouse events.

        (When the token list contained mouse handlers and the user clicked on
        on any of these, the matching handler is called. This handler can still
        return `NotImplemented` in case we want the `Window` to handle this
        particular event.)
        """
        if self._tokens:
            # Read the generator.
            tokens_for_line = list(split_lines(self._tokens))

            try:
                tokens = tokens_for_line[mouse_event.position.y]
            except IndexError:
                return NotImplemented
            else:
                # Find position in the token list.
                xpos = mouse_event.position.x

                # Find mouse handler for this character.
                count = 0
                for item in tokens:
                    count += len(item[1])
                    if count >= xpos:
                        if len(item) >= 3:
                            # Handler found. Call it.
                            # (Handler can return NotImplemented, so return
                            # that result.)
                            handler = item[2]
                            return handler(cli, mouse_event)
                        else:
                            break

        # Otherwise, don't handle here.
        return NotImplemented


class FillControl(UIControl):
    """
    Fill whole control with characters with this token.
    (Also helpful for debugging.)

    :param char: :class:`.Char` instance to use for filling.
    :param get_char: A callable that takes a CommandLineInterface and returns a
        :class:`.Char` object.
    """
    def __init__(self, character=None, token=Token, char=None, get_char=None):  # 'character' and 'token' parameters are deprecated.
        assert char is None or isinstance(char, Char)
        assert get_char is None or callable(get_char)
        assert not (char and get_char)

        self.char = char

        if character:
            # Passing (character=' ', token=token) is deprecated.
            self.character = character
            self.token = token

            self.get_char = lambda cli: Char(character, token)
        elif get_char:
            # When 'get_char' is given.
            self.get_char = get_char
        else:
            # When 'char' is given.
            self.char = self.char or Char()
            self.get_char = lambda cli: self.char
            self.char = char

    def __repr__(self):
        if self.char:
            return '%s(char=%r)' % (self.__class__.__name__, self.char)
        else:
            return '%s(get_char=%r)' % (self.__class__.__name__, self.get_char)

    def reset(self):
        pass

    def has_focus(self, cli):
        return False

    def create_content(self, cli, width, height):
        def get_line(i):
            return []

        return UIContent(
            get_line=get_line,
            line_count=100 ** 100,  # Something very big.
            default_char=self.get_char(cli))


_ProcessedLine = namedtuple('_ProcessedLine', 'tokens source_to_display display_to_source')


class BufferControl(UIControl):
    """
    Control for visualising the content of a `Buffer`.

    :param input_processors: list of :class:`~prompt_toolkit.layout.processors.Processor`.
    :param lexer: :class:`~prompt_toolkit.layout.lexers.Lexer` instance for syntax highlighting.
    :param preview_search: `bool` or `CLIFilter`: Show search while typing.
    :param get_search_state: Callable that takes a CommandLineInterface and
        returns the SearchState to be used. (If not CommandLineInterface.search_state.)
    :param buffer_name: String representing the name of the buffer to display.
    :param default_char: :class:`.Char` instance to use to fill the background. This is
        transparent by default.
    :param focus_on_click: Focus this buffer when it's click, but not yet focussed.
    """
    def __init__(self,
                 buffer_name=DEFAULT_BUFFER,
                 input_processors=None,
                 lexer=None,
                 preview_search=False,
                 search_buffer_name=SEARCH_BUFFER,
                 get_search_state=None,
                 menu_position=None,
                 default_char=None,
                 focus_on_click=False):
        assert input_processors is None or all(isinstance(i, Processor) for i in input_processors)
        assert menu_position is None or callable(menu_position)
        assert lexer is None or isinstance(lexer, Lexer)
        assert get_search_state is None or callable(get_search_state)
        assert default_char is None or isinstance(default_char, Char)

        self.preview_search = to_cli_filter(preview_search)
        self.get_search_state = get_search_state
        self.focus_on_click = to_cli_filter(focus_on_click)

        self.input_processors = input_processors or []
        self.buffer_name = buffer_name
        self.menu_position = menu_position
        self.lexer = lexer or SimpleLexer()
        self.default_char = default_char or Char(token=Token.Transparent)
        self.search_buffer_name = search_buffer_name

        #: Cache for the lexer.
        #: Often, due to cursor movement, undo/redo and window resizing
        #: operations, it happens that a short time, the same document has to be
        #: lexed. This is a faily easy way to cache such an expensive operation.
        self._token_cache = SimpleCache(maxsize=8)

        self._xy_to_cursor_position = None
        self._last_click_timestamp = None
        self._last_get_processed_line = None

    def _buffer(self, cli):
        """
        The buffer object that contains the 'main' content.
        """
        return cli.buffers[self.buffer_name]

    def has_focus(self, cli):
        # This control gets the focussed if the actual `Buffer` instance has the
        # focus or when any of the `InputProcessor` classes tells us that it
        # wants the focus. (E.g. in case of a reverse-search, where the actual
        # search buffer may not be displayed, but the "reverse-i-search" text
        # should get the focus.)
        return cli.current_buffer_name == self.buffer_name or \
            any(i.has_focus(cli) for i in self.input_processors)

    def preferred_width(self, cli, max_available_width):
        """
        This should return the preferred width.

        Note: We don't specify a preferred width according to the content,
              because it would be too expensive. Calculating the preferred
              width can be done by calculating the longest line, but this would
              require applying all the processors to each line. This is
              unfeasible for a larger document, and doing it for small
              documents only would result in inconsistent behaviour.
        """
        return None

    def preferred_height(self, cli, width, max_available_height, wrap_lines):
        # Calculate the content height, if it was drawn on a screen with the
        # given width.
        height = 0
        content = self.create_content(cli, width, None)

        # When line wrapping is off, the height should be equal to the amount
        # of lines.
        if not wrap_lines:
            return content.line_count

        # When the number of lines exceeds the max_available_height, just
        # return max_available_height. No need to calculate anything.
        if content.line_count >= max_available_height:
            return max_available_height

        for i in range(content.line_count):
            height += content.get_height_for_line(i, width)

            if height >= max_available_height:
                return max_available_height

        return height

    def _get_tokens_for_line_func(self, cli, document):
        """
        Create a function that returns the tokens for a given line.
        """
        # Cache using `document.text`.
        def get_tokens_for_line():
            return self.lexer.lex_document(cli, document)

        return self._token_cache.get(document.text, get_tokens_for_line)

    def _create_get_processed_line_func(self, cli, document):
        """
        Create a function that takes a line number of the current document and
        returns a _ProcessedLine(processed_tokens, source_to_display, display_to_source)
        tuple.
        """
        def transform(lineno, tokens):
            " Transform the tokens for a given line number. "
            source_to_display_functions = []
            display_to_source_functions = []

            # Get cursor position at this line.
            if document.cursor_position_row == lineno:
                cursor_column = document.cursor_position_col
            else:
                cursor_column = None

            def source_to_display(i):
                """ Translate x position from the buffer to the x position in the
                processed token list. """
                for f in source_to_display_functions:
                    i = f(i)
                return i

            # Apply each processor.
            for p in self.input_processors:
                transformation = p.apply_transformation(
                    cli, document, lineno, source_to_display, tokens)
                tokens = transformation.tokens

                if cursor_column:
                    cursor_column = transformation.source_to_display(cursor_column)

                display_to_source_functions.append(transformation.display_to_source)
                source_to_display_functions.append(transformation.source_to_display)

            def display_to_source(i):
                for f in reversed(display_to_source_functions):
                    i = f(i)
                return i

            return _ProcessedLine(tokens, source_to_display, display_to_source)

        def create_func():
            get_line = self._get_tokens_for_line_func(cli, document)
            cache = {}

            def get_processed_line(i):
                try:
                    # Try to get from cache first
                    return cache[i]
                except KeyError:
                    try:
                        # Not in cache, generate and store
                        processed_line = transform(i, get_line(i))
                        cache[i] = processed_line
                        return processed_line
                    except Exception:
                        # In case of any errors (like trying to access invalid line numbers),
                        # return an empty line instead of crashing
                        from rp.prompt_toolkit.token import Token
                        empty_line = transform(i, [(Token, '')])
                        cache[i] = empty_line
                        return empty_line
            return get_processed_line

        return create_func()

    def create_content(self, cli, width, height):
        """
        Create a UIContent.
        """
        buffer = self._buffer(cli)

        # Get the document to be shown. If we are currently searching (the
        # search buffer has focus, and the preview_search filter is enabled),
        # then use the search document, which has possibly a different
        # text/cursor position.)
        def preview_now():
            """ True when we should preview a search. """
            return bool(self.preview_search(cli) and
                        cli.buffers[self.search_buffer_name].text)

        if preview_now():
            if self.get_search_state:
                ss = self.get_search_state(cli)
            else:
                ss = cli.search_state

            document = buffer.document_for_search(SearchState(
                text=cli.current_buffer.text,
                direction=ss.direction,
                ignore_case=ss.ignore_case))
        else:
            document = buffer.document

        get_processed_line = self._create_get_processed_line_func(cli, document)
        self._last_get_processed_line = get_processed_line

        def translate_rowcol(row, col):
            " Return the content column for this coordinate. "
            return Point(y=row, x=get_processed_line(row).source_to_display(col))

        def get_line(i):
            " Return the tokens for a given line number. "
            tokens = get_processed_line(i).tokens

            # Add a space at the end, because that is a possible cursor
            # position. (When inserting after the input.) We should do this on
            # all the lines, not just the line containing the cursor. (Because
            # otherwise, line wrapping/scrolling could change when moving the
            # cursor around.)
            tokens = tokens + [(self.default_char.token, ' ')]
            return tokens

        content = UIContent(
            get_line=get_line,
            line_count=document.line_count,
            cursor_position=translate_rowcol(document.cursor_position_row,
                                             document.cursor_position_col),
            default_char=self.default_char)

        # If there is an auto completion going on, use that start point for a
        # pop-up menu position. (But only when this buffer has the focus --
        # there is only one place for a menu, determined by the focussed buffer.)
        if cli.current_buffer_name == self.buffer_name:
            menu_position = self.menu_position(cli) if self.menu_position else None
            if menu_position is not None:
                assert isinstance(menu_position, int)
                menu_row, menu_col = buffer.document.translate_index_to_position(menu_position)
                content.menu_position = translate_rowcol(menu_row, menu_col)
            elif buffer.complete_state:
                # Position for completion menu.
                # Note: We use 'min', because the original cursor position could be
                #       behind the input string when the actual completion is for
                #       some reason shorter than the text we had before. (A completion
                #       can change and shorten the input.)
                menu_row, menu_col = buffer.document.translate_index_to_position(
                    min(buffer.cursor_position,
                        buffer.complete_state.original_document.cursor_position))
                content.menu_position = translate_rowcol(menu_row, menu_col)
            else:
                content.menu_position = None

        return content

    def mouse_handler(self, cli, mouse_event):
        """
        Mouse handler for this control.
        """
        buffer = self._buffer(cli)
        position = mouse_event.position

        # Focus buffer when clicked.
        if self.has_focus(cli):
            if self._last_get_processed_line:
                processed_line = self._last_get_processed_line(position.y)

                # Translate coordinates back to the cursor position of the
                # original input.
                xpos = processed_line.display_to_source(position.x)
                index = buffer.document.translate_row_col_to_index(position.y, xpos)

                # Set the cursor position.
                if mouse_event.event_type == MouseEventType.MOUSE_DOWN:
                    buffer.exit_selection()
                    buffer.cursor_position = index

                elif mouse_event.event_type == MouseEventType.MOUSE_UP:
                    # When the cursor was moved to another place, select the text.
                    # (The >1 is actually a small but acceptable workaround for
                    # selecting text in Vi navigation mode. In navigation mode,
                    # the cursor can never be after the text, so the cursor
                    # will be repositioned automatically.)
                    if abs(buffer.cursor_position - index) > 1:
                        buffer.start_selection(selection_type=SelectionType.CHARACTERS)
                        buffer.cursor_position = index

                    # Select word around cursor on double click.
                    # Two MOUSE_UP events in a short timespan are considered a double click.
                    double_click = self._last_click_timestamp and time.time() - self._last_click_timestamp < .3
                    self._last_click_timestamp = time.time()

                    if double_click:
                        start, end = buffer.document.find_boundaries_of_current_word()
                        buffer.cursor_position += start
                        buffer.start_selection(selection_type=SelectionType.CHARACTERS)
                        buffer.cursor_position += end - start
                else:
                    # Don't handle scroll events here.
                    return NotImplemented

        # Not focussed, but focussing on click events.
        else:
            if self.focus_on_click(cli) and mouse_event.event_type == MouseEventType.MOUSE_UP:
                # Focus happens on mouseup. (If we did this on mousedown, the
                # up event will be received at the point where this widget is
                # focussed and be handled anyway.)
                cli.focus(self.buffer_name)
            else:
                return NotImplemented

    def move_cursor_down(self, cli):
        b = self._buffer(cli)
        b.cursor_position += b.document.get_cursor_down_position()

    def move_cursor_up(self, cli):
        b = self._buffer(cli)
        b.cursor_position += b.document.get_cursor_up_position()



# Stubs for sndhdr (Python 2 and 3)

import os
import sys
from typing import Any, NamedTuple, Optional, Tuple, Union

if sys.version_info >= (3, 5):
    SndHeaders = NamedTuple('SndHeaders', [
        ('filetype', str),
        ('framerate', int),
        ('nchannels', int),
        ('nframes', int),
        ('sampwidth', Union[int, str]),
    ])
    _SndHeaders = SndHeaders
else:
    _SndHeaders = Tuple[str, int, int, int, Union[int, str]]

if sys.version_info >= (3, 6):
    _Path = Union[str, bytes, os.PathLike[Any]]
else:
    _Path = Union[str, bytes]

def what(filename: _Path) -> Optional[_SndHeaders]: ...
def whathdr(filename: _Path) -> Optional[_SndHeaders]: ...



# Stubs for flask.logging (Python 3.6)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from .globals import request
from typing import Any

def wsgi_errors_stream(): ...
def has_level_handler(logger: Any): ...

default_handler: Any

def create_logger(app: Any): ...



import sys
from typing import Any, Callable, Dict, List, Optional, Tuple

if sys.version_info >= (3, 5):
    class JSONDecodeError(ValueError):
        msg: str
        doc: str
        pos: int
        lineno: int
        colno: int
        def __init__(self, msg: str, doc: str, pos: int) -> None: ...

class JSONDecoder:
    object_hook: Callable[[Dict[str, Any]], Any]
    parse_float: Callable[[str], Any]
    parse_int: Callable[[str], Any]
    parse_constant = ...  # Callable[[str], Any]
    strict: bool
    object_pairs_hook: Callable[[List[Tuple[str, Any]]], Any]

    def __init__(self, object_hook: Optional[Callable[[Dict[str, Any]], Any]] = ...,
                 parse_float: Optional[Callable[[str], Any]] = ...,
                 parse_int: Optional[Callable[[str], Any]] = ...,
                 parse_constant: Optional[Callable[[str], Any]] = ...,
                 strict: bool = ...,
                 object_pairs_hook: Optional[Callable[[List[Tuple[str, Any]]], Any]] = ...) -> None: ...
    def decode(self, s: str) -> Any: ...
    def raw_decode(self, s: str, idx: int = ...) -> Tuple[Any, int]: ...



from copy import deepcopy
from inspect import getfullargspec
from types import BuiltinFunctionType, BuiltinMethodType, FunctionType, MethodType

from typing import Any, Callable, Dict, Tuple, Type

from peepdis.preferences import _PreferencesMixin
from peepdis.datastructures import Arg, ArgDict, Output


# TODO: feature to check for state modifications by copying, then running method, then checking equality on attributes
# TODO: FIX for correct output from debug script
# TODO: should attrs and methods include builtins or should there be a
# TODO: verbose=False should exclude anything that's null, and builtins=False should exclude anything that's builtin. Verbose should be used to filter after eval
# TODO: ensure that forge is only forging positional args
# TODO: for data structures that contain multiple types, attempt to forge them, maybe using hypothesis strategies?


def peep(obj, verbose=False, forge=False):
    peeper = Peeper(obj)
    peeper.peep(forge)
    peeper.print(verbose)


class PeeperMixin(dict):
    def __init__(self, obj):
        super().__init__()
        self.obj = deepcopy(obj)
        self.dir = set(dir(obj))


class Peeper(PeeperMixin, _PreferencesMixin):
    def __init__(self, obj):
        # TODO: should I replace self with CallablePeeper?
        super().__init__(obj)
        self.attrs = dict()
        self.methods = dict()
        self.errors = dict()
        self.builtins = dict()

    def peep(self, forge=False):
        for name in self.dir:
            output = self.evaluate(name, forge=forge)
            self._index(output)

    def results(self, verbose=False):
        if verbose:
            return dict(self)
        else:
            return {name: self[name] for name in self if name not in self.builtins}

    def print(self, verbose=False):
        for name, output in self.results(verbose).items():
            print(*output.get_colored())

    def evaluate(self, name, *args, forge=False, **kwargs) -> Output:
        attr = getattr(self.obj, name)
        if not callable(attr):
            return Output(name, attr)
        # TODO: consider different paths for builtins
        unique_callables = (FunctionType, MethodType)
        builtin_callables = (BuiltinFunctionType, BuiltinMethodType)
        if isinstance(attr, unique_callables):
            peeper_class = CallablePeeper
        else:
            peeper_class = BuiltinCallablePeeper
        if name == "plot":
            a = 5
        try:
            peeper = peeper_class(attr)
            return peeper.evaluate(*args, forge=forge, **kwargs)
        except Exception as e:
            if "unsupported callable" in str(e):
                output_str = "(unsupported callable)"
                return Output(
                    name, output_str, callable_=True, unsupported_callable=True
                )
            else:
                raise ValueError(f"Unexpected error for method {name}")

    def _index(self, output: Output):
        self[output.name] = output
        if output.builtin:
            self.builtins[output.name] = self[output.name].value
        elif output.callable_:
            self.methods[output.name] = self[output.name].value
        elif output.error:
            self.errors[output.name] = self[output.name].value
        else:
            self.attrs[output.name] = self[output.name].value


class BuiltinCallablePeeper(PeeperMixin):
    # TODO: will need to merge with CallablePeeper in the future
    def __init__(self, obj, *args, **kwargs):
        super().__init__(obj)
        # TODO: move ArgDict instantiation up here and fix numbering

    def evaluate(self, *args, forge=False, **kwargs) -> Output:
        # TODO: figure out how to set up args in this case (no access to positional arg names)
        arg_names = [i for i in range(len(args))]
        arg_names += list(kwargs.values())
        self.args = ArgDict(arg_names)
        self.args.update_positional(*args)
        self.args.update_kwargs(**kwargs)
        obj = deepcopy(self.obj)
        try:
            output_str = repr(obj(*args, **kwargs))
        except Exception as e:
            output_str = repr(e)
            error = True
            if "argument" in str(e):
                args_missing = True
                if ":" in str(e):
                    missing_arg_str = output_str.split(":")[1][:-2]
                    output_str = f"(missing args: {missing_arg_str})"
                else:
                    output_str = f"(missing args)"
            else:
                args_missing = False
                output_str = repr(e)
        else:
            error = False
            args_missing = False
        return Output(
            obj.__name__,
            output_str,
            self.args,
            callable_=True,
            error=error,
            args_missing=args_missing,
        )


class CallablePeeper(PeeperMixin):
    _ordinary_callables = (FunctionType, MethodType)
    _builtin_callables = (BuiltinFunctionType, BuiltinMethodType)
    _forgery_dict = {"int": 0, "str": "abc", "list": [0, 1], "None": None}

    def __init__(self, obj, *args, **kwargs):
        super().__init__(obj)
        if isinstance(obj, self._ordinary_callables):
            self.spec = getfullargspec(obj)    # TODO: test on arrays, dataframes, and sklearn models (seemed to work on sklearn)
            self.args = ArgDict(self.spec.args)
            self._collect_args(*args, **kwargs)
        elif isinstance(obj, self._builtin_callables):
            self.args = ArgDict([])

    def _collect_args(self, *args, **kwargs):
        if self.spec.defaults:
            for i, arg in enumerate(self.spec.defaults[::-1]):
                self.args[-1 - i].value = arg
                self.args[-1 - i].method = "default"
        for i, arg in enumerate(args):
            self.args[i].value = arg
            self.args[i].method = "specified"
        for name, arg in kwargs.items():
            self.args[name].value = arg
            self.args[name].method = "specified"

    def evaluate(self, *args, forge=False, **kwargs) -> Output:
        obj = deepcopy(self.obj)
        if forge:
            self.forge_args()
        self.args.update_positional(*args)
        self.args.update_kwargs(**kwargs)
        arg_dict = self.args.generate_args()
        try:
            output_str = repr(obj(**arg_dict))
        except Exception as e:
            output_str = repr(e)
            error = True
            if "missing" in str(e) and "argument" in str(e):
                missing_arg_str = output_str.split(":")[1][:-2]
                output_str = f"(missing args: {missing_arg_str})"
                args_missing = True
            else:
                args_missing = False
                output_str = repr(e)
        else:
            error = False
            args_missing = False
        return Output(
            obj.__name__,
            output_str,
            self.args,
            callable_=True,
            error=error,
            args_missing=args_missing,
        )

    def forge_args(self, *args, **kwargs):
        """ Attempt to generate arguments to call the method """
        # TODO: currently forging kwargs and causing errors. Only forge positionals
        type_dict = self._infer_types()
        self.args.returns = type_dict.pop("returns", None)
        for name in self.args.null_args:
            type_ = type_dict.get(name, None)
            if not type_:
                # TODO: fill in with brute force
                continue
            self.args[name].value = self._forgery_dict[type_]
            self.args[name].method = "forged"
            self.args[name].type_ = type_
        for i, arg in enumerate(args):
            self.args[i].value = arg
            self.args[i].method = "specified"
        for name, arg in kwargs.items():
            self.args[name].value = arg
            self.args[name].method = "specified"
        return self.args

    def _infer_types(self, brute_force=True):
        """ Infer the argument types from the type annotations """
        # TODO: modify state of self.args rather than calling in forge_args
        type_dict = dict(self.spec.annotations)
        type_dict = {k: v.__name__ for k, v in type_dict}
        type_dict.pop("self", None)
        return type_dict

    def _brute_force_types(self):
        # TODO: iterate through missing arguments, trying different things
        # make a new class to do this?
        pass




"""Stub file for the '_heapq' module."""

from typing import TypeVar, List

_T = TypeVar("_T")

def heapify(heap: List[_T]) -> None: ...
def heappop(heap: List[_T]) -> _T:
    raise IndexError()  # if list is empty
def heappush(heap: List[_T], item: _T) -> None: ...
def heappushpop(heap: List[_T], item: _T) -> _T: ...
def heapreplace(heap: List[_T], item: _T) -> _T:
    raise IndexError()  # if list is empty
def nlargest(a: int, b: List[_T]) -> List[_T]: ...
def nsmallest(a: int, b: List[_T]) -> List[_T]: ...



from urwid import CompositeCanvas
from urwid import delegate_to_widget_mixin
from urwid import WidgetDecoration


class PopUpLauncher(delegate_to_widget_mixin('_original_widget'),
                    WidgetDecoration):
    """
    Re-implementation of :py:class:`urwid.PopUpLauncher` to add support for passing
    parameter when create/open pop-ups.

    This helps to create pop-ups with changing information.
    This reimplementation also pass size to :py:method:`.get_pup_up_parameters` when
    rendering, which helps to dynamically decide the position of the popups.
    """

    def __init__(self, original_widget):
        super().__init__(original_widget)
        self._pop_up_widget = None

    @property
    def pop_up_widget(self):
        return self._pop_up_widget

    def create_pop_up(self, *args, **kwargs):
        """
        Subclass must override this method and return a widget
        to be used for the pop-up.  This method is called once each time
        the pop-up is opened.
        """
        raise NotImplementedError("Subclass must override this method")

    def get_pop_up_parameters(self, *args, **kwargs):
        """
        Subclass must override this method and have it return a dict, eg:

        {'left':0, 'top':1, 'overlay_width':30, 'overlay_height':4}

        This method is called each time this widget is rendered.
        """
        raise NotImplementedError("Subclass must override this method")

    def open_pop_up(self, *args, **kwargs):
        self._pop_up_widget = self.create_pop_up(*args, **kwargs)
        self._invalidate()

    def close_pop_up(self, *args, **kwargs):
        self._pop_up_widget = None
        self._invalidate()

    def render(self, size, focus=False):
        canvas = super().render(size, focus)
        if self._pop_up_widget:
            canvas = CompositeCanvas(canvas)
            canvas.set_pop_up(
                self._pop_up_widget,
                **self.get_pop_up_parameters(size)
            )
        return canvas



#!/usr/bin/env python3.10
import sys
import re
from fstring_converter import convert_string

def normalize_code(code):
    """
    Normalize code for comparison by removing whitespace variations.
    This helps ensure that functionally equivalent code is considered equal
    even if whitespace or quotes differ slightly.
    """
    # Remove leading/trailing whitespace
    code = code.strip()
    
    # Normalize whitespace around operators
    code = re.sub(r'\s*([=+\-*/,])\s*', r'\1', code)
    
    # Normalize quotes - for test purposes, treating ' and " as equivalent
    # Not generally true, but helps with testing format strings
    def normalize_quotes(match):
        content = match.group(2)
        return f'"{content}"'
    
    code = re.sub(r"(['\"])(.*?)(\1)", normalize_quotes, code)
    
    return code

# Test cases covering all possible f-string scenarios
test_cases = [
    # Basic f-strings
    ('f"Hello, {name}"', '"Hello, {0}".format(name)'),
    ('f"{x} + {y} = {x + y}"', '"{0} + {1} = {2}".format(x, y, x + y)'),
    
    # Nested expressions
    ('f"Result: {calculate(x, y=3, z={"key": value})}"', 
     '"Result: {0}".format(calculate(x, y=3, z={"key": value}))'),
    
    # Format specifiers
    ('f"Pi: {math.pi:.2f}"', '"Pi: {0:.2f}".format(math.pi)'),
    ('f"Progress: {progress:.0%}"', '"Progress: {0:.0%}".format(progress)'),
    ('f"{num:05d}"', '"{0:05d}".format(num)'),
    
    # Conversions
    ('f"Debug: {obj!r}"', '"Debug: {0!r}".format(obj)'),
    ('f"String: {data!s}"', '"String: {0!s}".format(data)'),
    ('f"ASCII: {text!a}"', '"ASCII: {0!a}".format(text)'),
    
    # Expressions with format specifiers and conversions
    ('f"{complex_obj!r:>20}"', '"{0!r:>20}".format(complex_obj)'),
    
    # Escaped braces
    ('f"{{Not a variable}}"', '"{{Not a variable}}"'),
    ('f"{{These are literal braces}} but {this} is a variable"', 
     '"{{These are literal braces}} but {0} is a variable".format(this)'),
    
    # Raw f-strings
    ('rf"Raw string with {var}"', 'r"Raw string with {0}".format(var)'),
    
    # Multiline f-strings
    ('''f"""
    Multiline
    {variable}
    String
    """''', 
     '''"""
    Multiline
    {0}
    String
    """.format(variable)'''),
    
    # Complex expressions
    ('f"Math: {2 * (3 + 4)}"', '"Math: {0}".format(2 * (3 + 4))'),
    ('f"Dict: {data["key"]}"', '"Dict: {0}".format(data["key"])'),
    ('f"Method: {obj.method(arg1, arg2=val)}"', '"Method: {0}".format(obj.method(arg1, arg2=val))'),
    
    # Multiple expressions with different attributes
    ('f"{x.name} is {y[0].upper()}"', '"{0} is {1}".format(x.name, y[0].upper())'),
    
    # Empty f-string
    ('f""', '""'),
    
    # Multiple lines of code with f-strings
    ('''
def func():
    name = "Alice"
    age = 30
    print(f"{name} is {age} years old")
    return f"Info: {name}, {age}"
''', 
     '''
def func():
    name = "Alice"
    age = 30
    print("{0} is {1} years old".format(name, age))
    return "Info: {0}, {1}".format(name, age)
'''),
]

# Test real-world examples from actual codebase
real_world_cases = [
    ("""print(f"Saving {len(tensors)} tensors to {path}")""", 
     """print("Saving {0} tensors to {1}".format(len(tensors), path))"""),
    
    ("""print(f"    - {k}: {v.shape if hasattr(v, 'shape') else 'no shape'}")""", 
     """print("    - {0}: {1}".format(k, v.shape if hasattr(v, 'shape') else 'no shape'))"""),
]

# Add real-world test cases to our test suite
test_cases.extend(real_world_cases)

def run_tests():
    """Run all test cases and report failures"""
    passed = 0
    failed = 0
    errors = 0
    
    # First, try individual test cases
    for i, (input_code, expected_output) in enumerate(test_cases):
        try:
            actual_output = convert_string(input_code)
            
            # Function that evaluates if outputs are functionally equivalent
            def are_equivalent(actual, expected):
                # For simple strings, direct comparison
                if len(actual) < 100 and len(expected) < 100:
                    return normalize_code(actual) == normalize_code(expected)
                
                # For more complex code, check if key structures match
                # Strip all whitespace for comparison
                actual_condensed = re.sub(r'\s+', '', actual)
                expected_condensed = re.sub(r'\s+', '', expected)
                
                return actual_condensed == expected_condensed
            
            if are_equivalent(actual_output, expected_output):
                passed += 1
                print(f"✅ Test {i+1} passed")
            else:
                failed += 1
                print(f"❌ Test {i+1} failed")
                print(f"Input:    {input_code}")
                print(f"Expected: {expected_output}")
                print(f"Actual:   {actual_output}")
                print()
        except Exception as e:
            errors += 1
            print(f"❌ Test {i+1} failed with exception: {e}")
            print(f"Input:    {input_code}")
            print()
            
    # Real-world verification: The file from which this was extracted
    print("\n==== Testing Real File Conversion ====")
    try:
        file_test = """
def save_safetensors(tensors, path, metadata=None, *, verbose=False):
    \"\"\"
    Saves tensors to a .safetensors file.

    Args:
        tensors (dict or easydict): Dictionary of tensors to save
        path (str): Path to save the .safetensors file
        metadata (dict, optional): Metadata to include in the file. Defaults to None.
        verbose (bool, optional): Print tensor names. Defaults to False.

    Returns:
        str: Path to the saved file
        
    Reference: https://huggingface.co/docs/safetensors/en/index

    EXAMPLE:
        >>> tensors = {"weight": torch.randn(3, 4), "bias": torch.randn(4)}
        >>> save_safetensors(tensors, "model.safetensors")
        'model.safetensors'
        >>> loaded = load_safetensors("model.safetensors")
        >>> loaded.keys()
        ['weight', 'bias']
    \"\"\"
    import os
    from rp.r import pip_import
    pip_import("safetensors")
    from safetensors.torch import save_file
    
    # Convert to regular dict if it's an easydict
    if hasattr(tensors, 'to_dict'):
        tensors = tensors.to_dict()
    
    # Ensure all values are tensors
    if verbose:
        print(f"Saving {len(tensors)} tensors to {path}")
    
    for k, v in tensors.items():
        if verbose:
            print(f"    - {k}: {v.shape if hasattr(v, 'shape') else 'no shape'}")
    
    # Make sure the directory exists
    dir_path = os.path.dirname(path)
    if dir_path:
        os.makedirs(dir_path, exist_ok=True)
    
    # Save the tensors
    save_file(tensors, path, metadata)
    
    return path
"""
        converted = convert_string(file_test)
        
        # Check that f-strings are converted
        if "f\"" in converted or "f'" in converted:
            print("❌ File test failed - f-strings remain in the converted code")
            failed += 1
        else:
            print("✅ File test passed - all f-strings were converted")
            passed += 1
            
        # Visual check of conversion for debugging
        print("\nConverted real-world sample:")
        for line in converted.splitlines()[:20]:  # Show first 20 lines
            if ".format(" in line:
                print(f"  {line.strip()}")
    except Exception as e:
        print(f"❌ File test failed with exception: {e}")
        errors += 1
            
    print(f"\nResults: {passed} passed, {failed} failed, {errors} errors")
    return failed == 0 and errors == 0

if __name__ == "__main__":
    success = run_tests()
    sys.exit(0 if success else 1)